#!/usr/bin/env python

"""
Write sample QC results in 2 formats: a sample-lebel Hail Table,
and a TSV file.
"""

from os.path import join, splitext
import logging
import click
import hail as hl

from gnomad.utils.filtering import add_filters_expr

from joint_calling.utils import file_exists
from joint_calling import sample_qc as sqc, utils
from joint_calling import _version


logger = logging.getLogger(__file__)
logging.basicConfig(format='%(levelname)s (%(name)s %(lineno)s): %(message)s')
logger.setLevel(logging.INFO)


@click.command()
@click.version_option(_version.__version__)
@click.option(
    '--meta-csv',
    'meta_csv_path',
    required=True,
    help='path to a CSV with QC and population metadata for the samples',
)
@click.option(
    '--hard-filtered-samples-ht',
    'hard_filtered_samples_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
)
@click.option(
    '--hail-sample-qc-ht',
    'hail_sample_qc_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
)
@click.option(
    '--sex-ht',
    'sex_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
)
@click.option(
    '--custom-qc-ht',
    'custom_qc_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
)
@click.option(
    '--regressed-filtes-ht',
    'regressed_metrics_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
)
@click.option(
    '--relatedness-ht',
    'relatedness_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
)
@click.option(
    '--pop-ht',
    'pop_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
)
@click.option(
    '--filter-cutoffs-file',
    'filter_cutoffs_path',
    help=f'YAML file with filtering cutoffs',
)
@click.option(
    '--age-csv', 'age_csv_path', help='CSV file with 2 columns: `sample` and `age`'
)
@click.option(
    '--out-meta-ht',
    'out_meta_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=False),
)
@click.option(
    '--out-meta-tsv',
    'out_meta_tsv_path',
)
@click.option(
    '--tmp-bucket',
    'tmp_bucket',
    required=True,
    help='path to write temporary intermediate output and checkpoints',
)
@click.option(
    '--overwrite/--reuse',
    'overwrite',
    is_flag=True,
    help='if an intermediate or a final file exists, skip running the code '
    'that generates it.',
)
@click.option(
    '--hail-billing',
    'hail_billing',
    required=True,
    help='Hail billing account ID.',
)
def main(  # pylint: disable=too-many-arguments,too-many-locals,missing-function-docstring
    meta_csv_path: str,
    hard_filtered_samples_ht_path: str,
    hail_sample_qc_ht_path: str,
    sex_ht_path: str,
    custom_qc_ht_path: str,
    regressed_metrics_ht_path: str,
    relatedness_ht_path: str,
    pop_ht_path: str,
    filter_cutoffs_path: str,
    age_csv_path: str,
    out_meta_ht_path: str,
    out_meta_tsv_path: str,
    tmp_bucket: str,
    overwrite: bool,
    hail_billing: str,  # pylint: disable=unused-argument
):
    local_tmp_dir = utils.init_hail(__file__)

    cutoffs_d = utils.get_filter_cutoffs(filter_cutoffs_path)

    input_metadata_ht = utils.parse_input_metadata(meta_csv_path, local_tmp_dir)
    hard_filtered_samples_ht = hl.read_table(hard_filtered_samples_ht_path)
    sex_ht = hl.read_table(sex_ht_path)
    custom_qc_ht = hl.read_table(custom_qc_ht_path)
    hail_sample_qc_ht = hl.read_table(hail_sample_qc_ht_path)
    regressed_metrics_ht = hl.read_table(regressed_metrics_ht_path)
    relatedness_ht = hl.read_table(relatedness_ht_path)
    pop_ht = hl.read_table(pop_ht_path)

    if age_csv_path:
        age_ht = (
            hl.import_table(age_csv_path, delimiter=',', types={'age': 'float'})
            .rename({'TOBIID': 's'})
            .key_by('s')
        )
        input_metadata_ht = input_metadata_ht.annotate(
            age=age_ht[input_metadata_ht.external_id].age,
        )

    # Re-calculating the maximum set of unrelated samples now that
    # we have metrics adjusted for population, so newly QC-failed samples
    # are excluded
    related_samples_to_drop_ht_path = join(tmp_bucket, 'related_samples_to_drop.ht')
    related_samples_to_drop_ht = sqc.flag_related_samples(
        hard_filtered_samples_ht=hard_filtered_samples_ht,
        sex_ht=sex_ht,
        relatedness_ht=relatedness_ht,
        regressed_metrics_ht=regressed_metrics_ht,
        tmp_bucket=tmp_bucket,
        kin_threshold=cutoffs_d['pca']['max_kin'],
        out_ht_path=related_samples_to_drop_ht_path,
        overwrite=overwrite,
    )

    # Combine all intermediate tables together
    _generate_metadata(
        sample_qc_ht=hail_sample_qc_ht,
        custom_qc_ht=custom_qc_ht,
        sex_ht=sex_ht,
        input_metadata_ht=input_metadata_ht,
        hard_filtered_samples_ht=hard_filtered_samples_ht,
        regressed_metrics_ht=regressed_metrics_ht,
        pop_ht=pop_ht,
        related_samples_to_drop_after_qc_ht=related_samples_to_drop_ht,
        out_ht_path=out_meta_ht_path,
        out_tsv_path=out_meta_tsv_path,
        overwrite=overwrite,
    )


def _generate_metadata(
    sample_qc_ht: hl.Table,
    custom_qc_ht: hl.Table,
    sex_ht: hl.Table,
    input_metadata_ht: hl.Table,
    hard_filtered_samples_ht: hl.Table,
    regressed_metrics_ht: hl.Table,
    pop_ht: hl.Table,
    related_samples_to_drop_after_qc_ht: hl.Table,
    out_ht_path: str,
    out_tsv_path: str,
    overwrite: bool = False,
) -> hl.Table:
    """
    Combine all intermediate tables into a single metadata table

    :param sample_qc_ht: table with a `bi_allelic_sample_qc` row field
    :param custom_qc_ht: table with a `nongnomad_snps` row field
    :param sex_ht: table with the follwing row fields:
        `f_stat`, `n_called`, `expected_homs`, `observed_homs`
    :param input_metadata_ht: table with stats from the input metadata
        (includes Picard-tools stats and assigned population tags)
    :param hard_filtered_samples_ht: table with a `hard_filters` field
    :param regressed_metrics_ht: table with global fields
        `lms`, `qc_metrics_stats`;
        and metric row fields `n_snp_residual`, `r_ti_tv_residual`, etc;
        as well as corresponding `fail_n_snp_residual`, etc,
        and a `qc_metrics_filters` row field.
    :param pop_ht: table with the following row fields:
        `pop`, `prob_CEU`, `pca_scores`, `training_pop`.
    :param related_samples_to_drop_after_qc_ht:
        table with related samples, after re-ranking them based
        on population-stratified QC
    :param overwrite: overwrite checkpoints if they exist
    :return: table with relevant fields from input tables,
        annotated with the following row fields:
            'related_after_qc': bool = in `related_samples_to_drop_after_qc_ht`
            'related_before_qc': bool =
                in `related_samples_to_drop_before_qc_ht`
            'high_quality': bool =
                not `hard_filters_ht.hard_filters` &
                not `regressed_metrics_ht.qc_metrics_filters`
            'release': bool = high_quality & not `related`
    """
    logger.info('Generate the metadata HT and TSV files')

    if not overwrite and file_exists(out_ht_path):
        meta_ht = hl.read_table(out_ht_path)
    else:
        meta_ht = sex_ht.transmute(
            impute_sex_stats=hl.struct(
                f_stat=sex_ht.f_stat,
                n_called=sex_ht.n_called,
                expected_homs=sex_ht.expected_homs,
                observed_homs=sex_ht.observed_homs,
            )
        )

        meta_ht = meta_ht.annotate_globals(**regressed_metrics_ht.index_globals())

        meta_ht = meta_ht.annotate(
            sample_qc=sample_qc_ht[meta_ht.key].bi_allelic_sample_qc,
            nongnomad_snps=custom_qc_ht[meta_ht.key].nongnomad_snps,
            **hard_filtered_samples_ht[meta_ht.key],
            **regressed_metrics_ht[meta_ht.key],
            **pop_ht[meta_ht.key],
            **input_metadata_ht[meta_ht.key],
            related=hl.is_defined(related_samples_to_drop_after_qc_ht[meta_ht.key]),
        )

        meta_ht = meta_ht.annotate(
            hard_filters=hl.or_else(meta_ht.hard_filters, hl.empty_set(hl.tstr)),
            release_filters=add_filters_expr(
                filters={'related': meta_ht.related},
                current_filters=meta_ht.hard_filters.union(meta_ht.qc_metrics_filters),
            ),
        )

        meta_ht = meta_ht.annotate(
            high_quality=(hl.len(meta_ht.hard_filters) == 0),
            release=hl.len(meta_ht.release_filters) == 0,
        )
        meta_ht = meta_ht.checkpoint(out_ht_path, overwrite=True)

    out_tsv_path = out_tsv_path or splitext(out_ht_path)[0] + '.tsv'
    if overwrite or not file_exists(out_tsv_path):
        n_pcs = meta_ht.aggregate(hl.agg.min(hl.len(meta_ht.pca_scores)))
        meta_ht = meta_ht.transmute(
            **{f'PC{i + 1}': meta_ht.pca_scores[i] for i in range(n_pcs)},
            hard_filters=hl.or_missing(
                hl.len(meta_ht.hard_filters) > 0, hl.delimit(meta_ht.hard_filters)
            ),
            qc_metrics_filters=hl.or_missing(
                hl.len(meta_ht.qc_metrics_filters) > 0,
                hl.delimit(meta_ht.qc_metrics_filters),
            ),
        )
        meta_ht.flatten().export(out_tsv_path)

    return meta_ht


if __name__ == '__main__':
    main()  # pylint: disable=E1120



"""
Functions for variant QC plotting
"""

import logging
from typing import Set, Tuple, Union, Dict, List, Callable, Optional
from bokeh.layouts import gridplot
from bokeh.palettes import d3  # pylint: disable=no-name-in-module
from bokeh.models import (
    Plot,
    Row,
    Span,
    NumeralTickFormatter,
    LabelSet,
    Tabs,
    ColumnDataSource,
    DataRange1d,
    BooleanFilter,
    CDSView,
    HoverTool,
    Legend,
    Title,
    Panel,
    Column,
)
from bokeh.plotting import figure

import pandas as pd
import numpy as np
import hail as hl

from gnomad.utils.plotting import plot_hail_hist, plot_hail_hist_cumulative, TOOLS

logger = logging.getLogger('variant_qc_plotting')
logger.setLevel('INFO')


QC_PLOTS_SETTINGS = {
    'mean_point_size': 4.0,
    'min_point_size': 1.0,
    'max_point_size': 16.0,
    'label_text_font_size': '14pt',
    'title.text_font_size': '16pt',
    'subtitle.text_font_size': '14pt',
    'axis.axis_label_text_font_size': '16pt',
    'axis.axis_label_text_font_style': 'normal',
    'axis.major_label_text_font_size': '14pt',
}


def main():
    """
    :return:
    """
    pass


def get_point_size_col(data: pd.Series, size_prop: Optional[str]) -> pd.Series:
    """
    Given a data Series, returns the corresponding point size either:
    - Constant to qc_plots_settings['mean_point_size'] if `size_prop` is None
    - Radius proportional to data, if `size_prop` is 'radius'
    - Area proportional to data, if `size_prop` is 'area'

    Mean, min and max point  sizes are extracted from qc_plots_settings

    :param Series data: Input data series
    :param str size_prop: One of None, 'radius' or 'area'
    :return: Series with corresponding point size for each data point
    :rtype: Series
    """
    if size_prop is None:
        return pd.Series(len(data) * [QC_PLOTS_SETTINGS['mean_point_size']])

    mean_data = np.mean(data)
    if size_prop == 'radius':
        return data.apply(
            lambda x: max(
                QC_PLOTS_SETTINGS['min_point_size'],
                min(
                    QC_PLOTS_SETTINGS['max_point_size'],
                    QC_PLOTS_SETTINGS['mean_point_size'] * (x / mean_data),
                ),
            )
        )
    if size_prop == 'area':
        return data.apply(
            lambda x: max(
                QC_PLOTS_SETTINGS['min_point_size'],
                min(
                    QC_PLOTS_SETTINGS['max_point_size'],
                    QC_PLOTS_SETTINGS['mean_point_size']
                    * np.pi
                    * (np.sqrt(x / mean_data) / np.pi),
                ),
            )
        )
    raise ValueError(f'{size_prop} is not a supported value for argument `size_prop`')


def _set_plots_defaults(p: Plot) -> None:
    """
    :param p:
    :return:
    """
    p.legend.label_text_font_size = QC_PLOTS_SETTINGS['label_text_font_size']
    p.title.text_font_size = QC_PLOTS_SETTINGS['title.text_font_size']
    p.axis.axis_label_text_font_size = QC_PLOTS_SETTINGS[
        'axis.axis_label_text_font_size'
    ]
    p.axis.axis_label_text_font_style = QC_PLOTS_SETTINGS[
        'axis.axis_label_text_font_style'
    ]
    p.axis.major_label_text_font_size = QC_PLOTS_SETTINGS[
        'axis.major_label_text_font_size'
    ]


def get_binned_models_pd(
    score_ranking_binned_ht_by_model_name: Dict[str, str], contigs: Set[str] = None
) -> pd.DataFrame:
    """
    Creates a single DataFrame with all desired models binned and ready for plotting.

    :param score_ranking_binned_ht_by_model_name: a dict of model name matching
    a binned score ranking Table
    :param contigs: contigs to load
    :return: Plot-ready DataFrame
    :rtype: DataFrame
    """

    def aggregate_contig(ht: hl.Table, contigs: Set[str] = None):
        """
        Aggregates all contigs together and computes number for bins accross the contigs.
        """
        if contigs:
            ht = ht.filter(hl.literal(contigs).contains(ht.contig))

        return ht.group_by(*[k for k in ht.key if k != 'contig']).aggregate(
            min_score=hl.agg.min(ht.min_score),
            max_score=hl.agg.max(ht.max_score),
            **{
                x: hl.agg.sum(ht[x])
                for x in ht.row_value
                if x not in ['min_score', 'max_score']
            },
        )

    hts = [
        aggregate_contig(score_ht, contigs).annotate(model=model_name)
        for model_name, score_ht in score_ranking_binned_ht_by_model_name.items()
    ]

    ht = hts[0].union(*hts[1:])
    ht = ht.annotate(n_biallelic=hl.cond(ht.bi_allelic, ht.n, 0))
    return ht.to_pandas()


def plot_metric(
    df: pd.DataFrame,
    y_name: str,
    cols: List[str],
    y_fun: Callable[[pd.Series], Union[float, int]] = lambda x: x,
    cut: int = None,
    plot_all: bool = True,
    plot_bi_allelics: bool = True,
    plot_singletons: bool = True,
    plot_bi_allelic_singletons: bool = True,
    plot_adj: bool = False,
    colors: Dict[str, str] = None,
    link_cumul_y: bool = True,
    size_prop: str = 'area',
) -> Tabs:
    """
    Generic function for generating QC metric plots using a plotting-ready DataFrame (obtained from `get_binned_models_pd`)
    DataFrame needs to have a `rank_id` column, a `bin` column and a `model` column (contains the model name and needs to be added to binned table(s))

    This function generates scatter plots with the metric bin on x-axis and a user-defined function on the y-axis.
    The data for the y-axis function needs to from the columns specified in `cols`. The function is specified with the `y_fun` argument and data columns are access as a list.
    As an example, plotting Transition to transversion ratio is done as follows::

        plot_metric(snvs, 'Ti/Tv', ['n_ti', 'n_tv'], y_fun=lambda x: x[0]/x[1], colors=colors)

    In this command, `x[0]` correspond to the  first column selected (`'n_ti'`)  and `x[1]` to the second (`'n_tv'`).


    This function plots a tab for each of the plot condition(s) selected: all, bi-allelics, bi-allelic singletons.
    Within each tab, each row contains a non-cumulative and a cumulative plot of the bins / values.
    If `plot_adj` is set, then an extra row is added plotting only variants in release samples where AC_ADJ>0. The bin for these sites is computed based on those variants only.

    :param pd.DataFrame df: Input data
    :param str y_name: Name of the metric plotted on the y-axis
    :param list of str cols: Columns used to compute the metric plotted
    :param callable y_fun: Function to apply to the columns to generate the metric
    :param int cut: Where to draw the bin cut
    :param bool plot_all: Whether to plot a tab with all variants
    :param bool plot_bi_allelics: Whether to plot a tab with bi-allelic variants only
    :param bool plot_singletons: Whether to plot a tab with singleton variants only
    :param bool plot_bi_allelic_singletons:  Whether to plot a tab with bi-allelic singleton variants only
    :param bool plot_adj: Whether to plot additional rows with adj variants in release samples only
    :param dict of str -> str colors: Mapping of model name -> color
    :param bool link_cumul_y: If set, y-axes of cumulative and non-cumulative plots are linked
    :param str size_prop: Either 'size' or 'area' can be specified. If either is specified, the points will be sized proportionally to the amount of data in that point.
    :return: Plot
    :rtype: Tabs
    """

    def get_row(
        df: pd.DataFrame,
        y_name: str,
        cols: List[str],
        y_fun: Callable[[pd.Series], Union[float, int]],
        titles: List[str],
        link_cumul_y: bool,
        cut: int = None,
    ) -> Row:
        """
        Generates a single row with two plots: a regular scatter plot and a cumulative one.
        Both plots have bins on the x-axis. The y-axis is computed by applying the function `y_fun` on the columns `cols`.

        Data source is shared between the two plots so that highlighting / selection is linked.
        X-axis is shared between the two plots.
        Y-axus is shared if `link_cumul_y` is `True`

        """

        def get_plot(
            data_source: ColumnDataSource,
            y_name: str,
            y_col_name: str,
            titles: List[str],
            data_ranges: Tuple[DataRange1d, DataRange1d],
            cut: int = None,
        ) -> Plot:
            """
            Generates a single scatter plot panel
            """

            p = figure(
                title=titles[0],
                x_axis_label='bin',
                y_axis_label=y_name,
                tools='save,pan,box_zoom,reset,wheel_zoom,box_select,lasso_select,help,hover',
            )
            p.x_range = data_ranges[0]
            p.y_range = data_ranges[1]

            if cut:
                p.add_layout(
                    Span(
                        location=cut,
                        dimension='height',
                        line_color='red',
                        line_dash='dashed',
                    )
                )

            # Add circles layouts one model at a time, so that no default legend is generated.
            # Because data is in the same ColumnDataSource, use a BooleanFilter to plot each model separately
            circles = []
            for model in set(data_source.data['model']):
                view = CDSView(
                    source=data_source,
                    filters=[
                        BooleanFilter([x == model for x in data_source.data['model']])
                    ],
                )
                circles.append(
                    (
                        model,
                        [
                            p.circle(
                                'bin',
                                y_col_name,
                                color='_color',
                                size='_size',
                                source=data_source,
                                view=view,
                            )
                        ],
                    )
                )

            p.select_one(HoverTool).tooltips = [
                ('model', '@model'),
                ('bin', '@bin'),
                (y_name, f'@{y_col_name}'),
                ('min_score', '@min_score'),
                ('max_score', '@max_score'),
                ('n_data_points', '@_n'),
            ] + [(col, f'@{col}') for col in cols]
            _set_plots_defaults(p)

            # Add legend above the plot area
            legend = Legend(
                items=circles,
                orientation='horizontal',
                location=(0, 0),
                click_policy='hide',
            )
            p.add_layout(legend, 'above')

            # Add subtitles if any
            for title in titles[1:]:
                p.add_layout(
                    Title(
                        text=title,
                        text_font_size=QC_PLOTS_SETTINGS['subtitle.text_font_size'],
                    ),
                    'above',
                )

            return p

        # Compute non-cumulative values by applying `y_fun`
        df['non_cumul'] = df[cols].apply(y_fun, axis=1)

        # Compute cumulative values for each of the data columns
        for col in cols:
            df[f'{col}_cumul'] = df.groupby('model').aggregate(np.cumsum)[col]
        df['cumul'] = df[[f'{col}_cumul' for col in cols]].apply(y_fun, axis=1)

        # Create data ranges that are either shared or distinct depending on the y_cumul parameter
        non_cumul_data_ranges = (DataRange1d(), DataRange1d())
        cumul_data_ranges = (
            non_cumul_data_ranges
            if link_cumul_y
            else (non_cumul_data_ranges[0], DataRange1d())
        )
        data_source = ColumnDataSource(df)

        return Row(
            get_plot(
                data_source, y_name, 'non_cumul', titles, non_cumul_data_ranges, cut
            ),
            get_plot(
                data_source,
                y_name,
                'cumul',
                [titles[0] + ', cumulative'] + titles[1:],
                cumul_data_ranges,
                cut,
            ),
        )

    def prepare_pd(
        df: pd.DataFrame,
        cols: List[str],
        colors: Dict[str, str],
        size_prop: str = None,
    ):
        """
        Groups a pandas DataFrame by model and bin while keeping relevant columns only.
        Adds 3 columns used for plotting:
        1. A _color column column
        2. A _n column containing the number of data points
        3. A _size column containing the size of data points based on the `size_prop`
        and `qc_plot_settings` parameters
        """
        colors = colors or {}
        df = df.groupby(['model', 'bin']).agg(
            {**{col: np.sum for col in cols}, 'min_score': np.min, 'max_score': np.max}
        )
        df = df.reset_index()
        df['_color'] = [colors.get(x, 'gray') for x in df['model']]
        df['_n'] = np.sum(df[cols], axis=1)
        df['_size'] = get_point_size_col(df['_n'], size_prop)
        return df

    colors = colors if colors is not None else {}
    tabs = []
    adj_strats = ['', 'adj_'] if plot_adj else ['']

    if plot_all:
        children = []
        for adj in adj_strats:
            titles = [y_name, 'Adj variants (adj rank)' if adj else 'All variants']
            plot_df = prepare_pd(
                df.loc[df.rank_id == f'{adj}rank'], cols, colors, size_prop
            )
            if len(plot_df) > 0:
                children.append(
                    get_row(plot_df, y_name, cols, y_fun, titles, link_cumul_y, cut)
                )
            else:
                logger.warning('No data found for plot: ' + '\t'.join(titles))

        if children:
            tabs.append(Panel(child=Column(children=children), title='All'))

    if plot_bi_allelics:
        children = []
        for adj in adj_strats:
            for biallelic_rank in ['', 'biallelic_']:
                titles = [
                    y_name,
                    'Bi-allelic variants ({} rank)'.format(
                        'overall'
                        if not adj and not biallelic_rank
                        else ' '.join([adj[:-1], biallelic_rank[:-1]]).lstrip()
                    ),
                ]
                plot_df = prepare_pd(
                    df.loc[
                        df.bi_allelic & (df.rank_id == f'{adj}{biallelic_rank}rank')
                    ],
                    cols,
                    colors,
                    size_prop,
                )
                if len(plot_df) > 0:
                    children.append(
                        get_row(plot_df, y_name, cols, y_fun, titles, link_cumul_y, cut)
                    )
                else:
                    logger.warning('No data found for plot: ' + '\t'.join(titles))

        if children:
            tabs.append(Panel(child=Column(children=children), title='Bi-allelic'))

    if plot_singletons:
        children = []
        for adj in adj_strats:
            for singleton_rank in ['', 'singleton_']:
                titles = [
                    y_name,
                    'Singletons ({} rank)'.format(
                        'overall'
                        if not adj and not singleton_rank
                        else ' '.join([adj[:-1], singleton_rank[:-1]]).lstrip()
                    ),
                ]
                plot_df = prepare_pd(
                    df.loc[df.singleton & (df.rank_id == f'{adj}{singleton_rank}rank')],
                    cols,
                    colors,
                    size_prop,
                )
                if len(plot_df) > 0:
                    children.append(
                        get_row(plot_df, y_name, cols, y_fun, titles, link_cumul_y, cut)
                    )
                else:
                    logger.warning('No data found for plot: ' + '\t'.join(titles))

        if children:
            tabs.append(Panel(child=Column(children=children), title='Singletons'))

    if plot_bi_allelic_singletons:
        children = []
        for adj in adj_strats:
            for bisingleton_rank in ['', 'biallelic_singleton_']:
                titles = [
                    y_name,
                    'Bi-allelic singletons ({} rank)'.format(
                        'overall'
                        if not adj and not bisingleton_rank
                        else ' '.join(
                            [adj[:-1], bisingleton_rank[:-1].replace('_', ' ')]
                        ).lstrip()
                    ),
                ]
                plot_df = prepare_pd(
                    df.loc[
                        df.bi_allelic
                        & df.singleton
                        & (df.rank_id == f'{adj}{bisingleton_rank}rank')
                    ],
                    cols,
                    colors,
                    size_prop,
                )
                if len(plot_df) > 0:
                    children.append(
                        get_row(plot_df, y_name, cols, y_fun, titles, link_cumul_y, cut)
                    )
                else:
                    logger.warning('No data found for plot: ' + '\t'.join(titles))

        if children:
            tabs.append(
                Panel(child=Column(children=children), title='Bi-allelic singletons')
            )

    return Tabs(tabs=tabs)


def plot_score_distributions(
    model_name_by_model_id: Dict[str, str],
    score_ranking_path_by_model_id: Dict[str, str],
    binned_score_ranking_path_by_model_id: Dict[str, str],
    snv: bool,
    cut: int,
    colors: Dict[str, str] = None,
) -> Tabs:
    """
    Generates plots of model scores distributions:
    One tab per model.
    Within each tab, there is 2x2 grid of plots:
    - One row showing the score distribution across the entire data
    - One row showing the score distribution across the release-samples,
      adj data only (release_sample_AC_ADJ > 0)
    - One column showing the histogram of the score
    - One column showing the normalized cumulative histogram of the score

    Cutoff is highlighted by a dashed red line

    :param model_name_by_model_id:
           Which models to plot. A dict with mapping from model id
           to model name for display.
    :param score_ranking_path_by_model_id: unbinned score ranking Table paths,
           or RF result Table for the RF model
    :param binned_score_ranking_path_by_model_id: binned score ranking Table paths
    :param bool snv: Whether to plot SNVs or Indels
    :param int cut: Bin cut on the entire data to highlight
    :param dict of str -> str colors: Optional colors to use (model name -> desired color)
    :return: Plots of the score distributions
    :rtype: Tabs
    """

    if colors is None:
        colors = {m_name: '#033649' for m_name in model_name_by_model_id.values()}

    tabs = []
    for model_id, model_name in model_name_by_model_id.items():
        # if model_id in ['vqsr', 'cnn', 'rf_2.0.2', 'rf_2.0.2_beta']:
        ht = hl.read_table(score_ranking_path_by_model_id[model_id])

        ht = ht.filter(hl.is_snp(ht.alleles[0], ht.alleles[1]), keep=snv)
        binned_ht = hl.read_table(binned_score_ranking_path_by_model_id[model_id])
        binned_ht = binned_ht.filter(binned_ht.snv, keep=snv)

        cut_value = binned_ht.aggregate(
            hl.agg.filter(
                (binned_ht.bin == cut) & (binned_ht.rank_id == 'rank'),
                hl.agg.min(binned_ht.min_score),
            )
        )

        min_score, max_score = (-20, 30) if model_id == 'vqsr' else (0.0, 1.0)
        agg_values = ht.aggregate(
            hl.struct(
                score_hist=[
                    hl.agg.hist(ht.score, min_score, max_score, 100),
                    hl.agg.filter(
                        ht.ac > 0, hl.agg.hist(ht.score, min_score, max_score, 100)
                    ),
                ],
                adj_counts=hl.agg.filter(
                    ht.ac > 0, hl.agg.counter(ht.score >= cut_value)
                ),
            )
        )
        score_hist = agg_values.score_hist
        adj_cut = '{0:.2f}'.format(
            100
            * agg_values.adj_counts[True]
            / (agg_values.adj_counts[True] + agg_values.adj_counts[False])
        )

        rows = []
        x_range = DataRange1d()
        y_range = [DataRange1d(), DataRange1d()]
        for adj in [False, True]:
            title = '{0}, {1} cut (score = {2:.2f})'.format(
                'Adj' if adj else 'All', adj_cut if adj else cut, cut_value
            )
            p = plot_hail_hist(
                score_hist[adj], title=title + '\n', fill_color=colors[model_name]
            )
            p.add_layout(
                Span(
                    location=cut_value,
                    dimension='height',
                    line_color='red',
                    line_dash='dashed',
                )
            )
            p.x_range = x_range
            p.y_range = y_range[0]
            _set_plots_defaults(p)

            p_cumul = plot_hail_hist_cumulative(
                score_hist[adj],
                title=title + ', cumulative',
                line_color=colors[model_name],
            )
            p_cumul.add_layout(
                Span(
                    location=cut_value,
                    dimension='height',
                    line_color='red',
                    line_dash='dashed',
                )
            )
            p_cumul.x_range = x_range
            p_cumul.y_range = y_range[1]
            _set_plots_defaults(p_cumul)

            rows.append([p, p_cumul])

        tabs.append(Panel(child=gridplot(rows), title=model_name))
    return Tabs(tabs=tabs)


def get_binned_concordance_pd(
    model_name_by_model_id: Dict[str, str],
    binned_concordance_ht_path_by_model_id: Dict[str, str],
    truth_samples: List[str],
) -> pd.DataFrame:
    """
    Creates a pandas DF containing the binned concordance results for all given truth
     samples / models.

    :param model_name_by_model_id:
           Which models to plot. A dict with mapping from model id
           to model name for display.
    :param list of str truth_samples: List of truth samples to include
    :return: Pandas dataframe with binned concordance results
    :rtype: DataFrame
    """

    def get_binned_concordance_ht(
        truth_samples: List[str],
        model_name_by_model_id: Dict[str, str],
    ) -> hl.Table:
        """
        Combines binned concordance results for multiple truth samples and/or models
        into a single Table.
        """
        hts = []
        for truth_sample in truth_samples:
            for model_id, model_name in model_name_by_model_id.items():
                conc_path = binned_concordance_ht_path_by_model_id[model_id]
                # conc_path = join(
                #     work_bucket, f'binned_concordance/{truth_sample}_{model_id}.ht'
                # )
                ht = hl.read_table(conc_path)
                ht = ht.annotate(truth_sample=truth_sample, model=model_name)
                hts.append(ht)

        return hts[0].union(*hts[1:])

    def compute_cumul_metrics(df: pd.DataFrame) -> pd.DataFrame:
        """
        Computes cumulative metrics on a pandas DF.
        """
        df = df.sort_values(by=['bin'])
        df['cum_tp'] = df['tp'].cumsum()
        df['cum_fp'] = df['fp'].cumsum()
        total_pos = df['tp'].sum() + df['fn'].sum()
        total_neg = df['fp'].sum()
        df['cum_tn'] = total_neg - df['cum_fp']
        df['cum_fn'] = total_pos - df['cum_tp']
        df['precision'] = df['cum_tp'] / (df['cum_tp'] + df['cum_fp'])
        df['recall'] = df['cum_tp'] / (df['cum_tp'] + df['cum_fn'])
        df['cum_alleles'] = df['n_alleles'].cumsum()
        return df[
            [
                'bin',
                'min_score',
                'max_score',
                'n_alleles',
                'tp',
                'fp',
                'fn',
                'cum_alleles',
                'cum_tp',
                'cum_fp',
                'cum_fn',
                'cum_tn',
                'precision',
                'recall',
            ]
        ]

    df = get_binned_concordance_ht(truth_samples, model_name_by_model_id).to_pandas()
    df = df.groupby(['rank_name', 'truth_sample', 'model', 'snv']).apply(
        compute_cumul_metrics
    )
    return df.fillna(-1).groupby(['rank_name', 'truth_sample', 'model', 'snv'])


def plot_concordance_pr(
    pr_df: pd.DataFrame,
    snv: bool,
    colors: Dict[str, str] = None,
    size_prop: str = None,
    bins_to_label: List[int] = None,
) -> Column:
    """
    Generates plots showing Precision/Recall curves for truth samples:
    Two tabs:
    - One displaying the PR curve with ranking computed on the entire data
    - One displaying the PR curve with ranking computed on the  truth sample only

    Within each tab, a row of n_truth_samples.

    The input to this function should come out of the `get_binned_concordance_pd` function, which creates
    a DataFrame containing the necessary metris for PR plotting and is grouped by 'rank_name', 'truth_sample', 'model' and 'snv'.

    :param DataFrame pr_df: Input Dataframe
    :param bool snv: Whether to plot SNVs or Indels
    :param dict of str -> str colors: Optional colors to use (model name -> desired color)
    :param str size_prop: Either 'radius' or 'area' can be specified. If either is specified, the points will be sized proportionally to the amount of data in that point.
    :param list of int bins_to_label: Bins to label
    :return: Bokeh grid of plots
    :rtype: Tabs
    """

    if colors is None:
        # Get a palette automatically
        models = sorted(list(set([g[2] for g in pr_df.groups])))
        palette = d3['Category10'][max(3, len(models))]
        colors = {model: palette[i] for i, model in enumerate(models)}

    hover = HoverTool(
        tooltips=[
            ('model', '@model'),
            ('bin', '@bin'),
            ('score (min, max)', '(@min_score, @max_score)'),
            ('n_alleles', '@n_alleles'),
            ('cum_alleles', '@cum_alleles'),
            ('data (x,y)', '($x, $y)'),
        ]
    )

    tabs = []
    for rank in ['truth_sample_rank', 'global_rank']:

        plot_row = []
        for truth_sample in set([g[1] for g in pr_df.groups]):

            p = figure(
                title=truth_sample[0].upper() + truth_sample[1:],
                x_axis_label='Recall',
                y_axis_label='Precision',
                tools=[hover] + [tool for tool in TOOLS.split(',') if tool != 'hover'],
            )
            p.xaxis[0].formatter = NumeralTickFormatter(format='0%')
            p.yaxis[0].formatter = NumeralTickFormatter(format='0.0%')

            circles = []
            for model in set([g[2] for g in pr_df.groups]):
                data = pr_df.get_group((rank, truth_sample, model, snv)).copy()
                data['model'] = [model] * len(data)
                data['size'] = get_point_size_col(data['n_alleles'], size_prop)
                source = ColumnDataSource(data)
                circles.append(
                    (
                        model,
                        [
                            p.circle(
                                'recall',
                                'precision',
                                size='size',
                                color=colors[model],
                                source=source,
                            )
                        ],
                    )
                )
                if bins_to_label is not None:
                    label_data = data.loc[data.bin.isin(bins_to_label)].copy()
                    label_data['x_offset'] = label_data['recall'] + 0.025
                    label_data['y_offset'] = label_data['precision']
                    label_data['bin_str'] = [str(int(t)) for t in label_data['bin']]
                    label_source = ColumnDataSource(label_data)
                    p.add_layout(
                        LabelSet(
                            x='x_offset',
                            y='precision',
                            text='bin_str',
                            text_color=colors[model],
                            source=label_source,
                        )
                    )
                    p.multi_line(
                        xs=[[x, x + 0.05] for x in label_data.recall],
                        ys=[[y, y] for y in label_data.precision],
                        color=colors[model],
                    )

            legend = Legend(
                items=circles,
                orientation='horizontal',
                location=(0, 0),
                click_policy='hide',
            )
            p.add_layout(legend, 'above')
            _set_plots_defaults(p)
            plot_row.append(p)

        tabs.append(Panel(child=Row(children=plot_row), title=rank))

    return Tabs(tabs=tabs)


if __name__ == '__main__':
    main()  # pylint: disable=E1120



#!/usr/bin/env python

"""
Combine a set of GVCFs into a MatrixTable
"""

import os
from os.path import join, basename
import subprocess
from typing import List
import logging
import shutil

import pandas as pd
import click
import hail as hl
from hail.experimental.vcf_combiner import vcf_combiner

from joint_calling.utils import get_validation_callback
from joint_calling import utils
from joint_calling import _version

logger = logging.getLogger('combine_gvcfs')
logger.setLevel('INFO')

DEFAULT_REF = 'GRCh38'
# The target number of rows per partition during each round of merging
TARGET_RECORDS = 25_000


@click.command()
@click.version_option(_version.__version__)
@click.option(
    '--meta-csv',
    'meta_csv_path',
    required=True,
    help='Sample data CSV path',
)
@click.option(
    '--out-mt',
    'out_mt_path',
    required=True,
    callback=get_validation_callback(ext='mt'),
    help='path to write the combined MatrixTable',
)
@click.option(
    '--existing-mt',
    'existing_mt_path',
    callback=get_validation_callback(ext='mt', must_exist=True),
    help='optional path to an existing MatrixTable. '
    'If provided, will be read and used as a base to get extended with the '
    'samples in the input sample map. Can be read-only, as it will not '
    'be overwritten, instead the result will be written to the new location '
    'provided with --out-mt',
)
@click.option(
    '--bucket',
    'work_bucket',
    required=True,
    help='path to folder for intermediate output. '
    'Can be a Google Storage URL (i.e. start with `gs://`).',
)
@click.option(
    '--local-tmp-dir',
    'local_tmp_dir',
    help='local directory for temporary files and Hail logs (must be local).',
)
@click.option(
    '--overwrite/--reuse',
    'overwrite',
    is_flag=True,
    help='if an intermediate or a final file exists, skip running the code '
    'that generates it.',
)
@click.option(
    '--hail-billing',
    'hail_billing',
    help='Hail billing account ID.',
)
@click.option(
    '--n-partitions',
    'n_partitions',
    type=click.INT,
    help='Number of partitions for the output matrix table',
)
def main(
    meta_csv_path: str,
    out_mt_path: str,
    existing_mt_path: str,
    work_bucket: str,
    local_tmp_dir: str,
    overwrite: bool,  # pylint: disable=unused-argument
    hail_billing: str,  # pylint: disable=unused-argument
    n_partitions: int,
):  # pylint: disable=missing-function-docstring
    local_tmp_dir = utils.init_hail('combine_gvcfs', local_tmp_dir)

    logger.info(f'Combining GVCFs')

    assert utils.file_exists(meta_csv_path)
    local_meta_csv_path = join(local_tmp_dir, basename(meta_csv_path))
    subprocess.run(
        f'gsutil cp {meta_csv_path} {local_meta_csv_path}', check=False, shell=True
    )
    new_samples_df = pd.read_table(local_meta_csv_path)

    new_mt_path = os.path.join(work_bucket, 'new.mt')
    combine_gvcfs(
        gvcf_paths=list(new_samples_df.gvcf),
        sample_names=list(new_samples_df.s),
        out_mt_path=new_mt_path,
        work_bucket=work_bucket,
        overwrite=True,
    )
    new_mt = hl.read_matrix_table(new_mt_path)
    logger.info(
        f'Written {new_mt.cols().count()} new samples to {out_mt_path}, '
        f'n_partitions={new_mt.n_partitions()}'
    )

    new_plus_existing_mt = None
    if existing_mt_path:
        logger.info(f'Combining with the existing matrix table {existing_mt_path}')
        new_plus_existing_mt = _combine_with_the_existing_mt(
            existing_mt_path=existing_mt_path,
            new_mt_path=new_mt_path,
        )

    mt = new_plus_existing_mt or new_mt
    mt.repartition(n_partitions)
    mt.write(out_mt_path, overwrite=True)
    logger.info(
        f'Written {mt.count_cols()} samples to {out_mt_path}, '
        f'n_partitions={mt.n_partitions()}'
    )

    shutil.rmtree(local_tmp_dir)


def _combine_with_the_existing_mt(
    existing_mt_path: str,
    # passing as a path because we are going
    # to re-read it with different intervals
    new_mt_path: str,
) -> hl.MatrixTable:
    # Making sure the tables are keyed by locus only, to make the combiner work.
    existing_mt = hl.read_matrix_table(existing_mt_path).key_rows_by('locus')
    intervals = vcf_combiner.calculate_new_intervals(
        existing_mt.rows(),
        n=TARGET_RECORDS,
        reference_genome=DEFAULT_REF,
    )
    new_mt = hl.read_matrix_table(new_mt_path, _intervals=intervals).key_rows_by(
        'locus'
    )
    logger.info(
        f'Combining {new_mt_path} ({new_mt.count_cols()} samples) '
        f'with an existing MatrixTable {existing_mt_path} '
        f'({existing_mt.count_cols()} samples), '
        f'split into {existing_mt.n_partitions()} partitions)'
    )
    out_mt = vcf_combiner.combine_gvcfs([existing_mt, new_mt])
    return out_mt


def combine_gvcfs(
    gvcf_paths: List[str],
    sample_names: List[str],
    out_mt_path: str,
    work_bucket: str,
    overwrite: bool = True,
):
    """
    Combine a set of GVCFs in one go
    """
    hl.experimental.run_combiner(
        gvcf_paths,
        sample_names=sample_names,
        out_file=out_mt_path,
        reference_genome=utils.DEFAULT_REF,
        use_genome_default_intervals=True,
        tmp_path=os.path.join(work_bucket, 'tmp'),
        overwrite=overwrite,
        key_by_locus_and_alleles=True,
    )


if __name__ == '__main__':
    main()  # pylint: disable=E1120



#!/usr/bin/env python

"""
Plot ancestry PCA analysis results
"""

import logging
import click
import pandas as pd
import numpy as np
import hail as hl
from bokeh.io.export import get_screenshot_as_png
from bokeh.resources import CDN
from bokeh.embed import file_html
from bokeh.transform import factor_cmap
from bokeh.plotting import ColumnDataSource, figure
from bokeh.palettes import turbo  # pylint: disable=no-name-in-module
from bokeh.models import CategoricalColorMapper, HoverTool

from joint_calling import utils
from joint_calling import _version


logger = logging.getLogger(__file__)
logging.basicConfig(format='%(levelname)s (%(name)s %(lineno)s): %(message)s')
logger.setLevel(logging.INFO)


@click.command()
@click.version_option(_version.__version__)
@click.option(
    '--eigenvalues',
    'eigenvalues_path',
    required=True,
    callback=utils.get_validation_callback(ext='txt', must_exist=True),
)
@click.option(
    '--scores-ht',
    'scores_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
)
@click.option(
    '--loadings-ht',
    'loadings_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
)
@click.option(
    '--provided-pop-ht',
    'provided_pop_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
)
@click.option(
    '--inferred-pop-ht',
    'inferred_pop_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
)
@click.option('--out-path-pattern', 'out_path_pattern')
@click.option(
    '--hail-billing',
    'hail_billing',
    required=True,
    help='Hail billing account ID.',
)
def main(  # pylint: disable=too-many-arguments,too-many-locals,missing-function-docstring
    eigenvalues_path: str,
    scores_ht_path: str,
    loadings_ht_path: str,
    provided_pop_ht_path: str,
    inferred_pop_ht_path: str,
    out_path_pattern: str,
    hail_billing: str,  # pylint: disable=unused-argument
):
    utils.init_hail(__file__)

    produce_plots(
        eigenvalues_path=eigenvalues_path,
        scores_ht_path=scores_ht_path,
        loadings_ht_path=loadings_ht_path,
        provided_pop_ht_path=provided_pop_ht_path,
        inferred_pop_ht_path=inferred_pop_ht_path,
        out_path_pattern=out_path_pattern,
    )


def produce_plots(
    eigenvalues_path: str,
    scores_ht_path: str,
    loadings_ht_path: str,
    provided_pop_ht_path: str,
    inferred_pop_ht_path: str,
    out_path_pattern: str,
):
    """
    Generate plots in png and html formats, write for each PC (of n_pcs) and
    scope ("study", "continental_pop", "subpop", plus for loadings) into
    file paths defined by `out_path_pattern`.
    """
    scores = hl.read_table(scores_ht_path)
    sample_names = scores.s.collect()
    provided_pop_ht = hl.read_table(provided_pop_ht_path)
    inferred_pop_ht = hl.read_table(inferred_pop_ht_path)
    scores = scores.annotate(
        continental_pop=hl.case()
        .when(
            provided_pop_ht[scores.s].continental_pop != '',
            provided_pop_ht[scores.s].continental_pop,
        )
        .default(inferred_pop_ht[scores.s].pop + ' [inferred]'),
        subpop=hl.case()
        .when(provided_pop_ht[scores.s].subpop != '', provided_pop_ht[scores.s].subpop)
        .default(inferred_pop_ht[scores.s].pop + ' [inferred]'),
        study=provided_pop_ht[scores.s].project,
    )

    # plot by study
    labels = scores.study.collect()
    study = list(set(labels))
    tooltips = [('labels', '@label'), ('samples', '@samples')]
    eigenvalues = hl.import_table(
        eigenvalues_path, no_header=True, types={'f0': hl.tfloat}
    ).f0.collect()
    eigenvalues = pd.to_numeric(eigenvalues)
    variance = np.divide(eigenvalues[1:], float(eigenvalues.sum())) * 100
    variance = variance.round(2)

    # Get number of PCs
    number_of_pcs = len(eigenvalues) - 1

    for i in range(number_of_pcs - 1):
        pc1 = i
        pc2 = i + 1
        plot = figure(
            title='Study',
            x_axis_label=f'PC{pc1 + 1} ({variance[pc1]})%)',
            y_axis_label=f'PC{pc2 + 1} ({variance[pc2]}%)',
            tooltips=tooltips,
        )
        source = ColumnDataSource(
            dict(
                x=scores.scores[pc1].collect(),
                y=scores.scores[pc2].collect(),
                label=labels,
                samples=sample_names,
            )
        )
        plot.circle(
            'x',
            'y',
            alpha=0.5,
            source=source,
            size=4,
            color=factor_cmap('label', ['#1b9e77', '#d95f02'], study),
            legend_group='label',
        )
        plot.add_layout(plot.legend[0], 'left')
        plot_filename = out_path_pattern.format(scope='study', pci=pc2, ext='png')
        with hl.hadoop_open(plot_filename, 'wb') as f:
            get_screenshot_as_png(plot).save(f, format='PNG')
        html = file_html(plot, CDN, 'my plot')
        plot_filename_html = out_path_pattern.format(scope='study', pci=pc2, ext='html')
        with hl.hadoop_open(plot_filename_html, 'w') as f:
            f.write(html)

    # plot by continental population
    labels = scores.continental_pop.collect()
    continental_population = list(set(labels))
    tooltips = [('labels', '@label'), ('samples', '@samples')]

    for i in range(number_of_pcs - 1):
        pc1 = i
        pc2 = i + 1
        plot = figure(
            title='Continental Population',
            x_axis_label=f'PC{pc1 + 1} ({variance[pc1]})%)',
            y_axis_label=f'PC{pc2 + 1} ({variance[pc2]}%)',
            tooltips=tooltips,
        )
        source = ColumnDataSource(
            dict(
                x=scores.scores[pc1].collect(),
                y=scores.scores[pc2].collect(),
                label=labels,
                samples=sample_names,
            )
        )
        plot.circle(
            'x',
            'y',
            alpha=0.5,
            source=source,
            size=4,
            color=factor_cmap(
                'label', turbo(len(continental_population)), continental_population
            ),
            legend_group='label',
        )
        plot.add_layout(plot.legend[0], 'left')
        plot_filename = out_path_pattern.format(
            scope='continental_pop', pci=pc2, ext='png'
        )
        with hl.hadoop_open(plot_filename, 'wb') as f:
            get_screenshot_as_png(plot).save(f, format='PNG')
        html = file_html(plot, CDN, 'my plot')
        plot_filename_html = out_path_pattern.format(
            scope='continental_pop', pci=pc2, ext='html'
        )
        with hl.hadoop_open(plot_filename_html, 'w') as f:
            f.write(html)

    # plot by subpopulation
    labels = scores.subpop.collect()
    sub_population = list(set(labels))
    tooltips = [('labels', '@label'), ('samples', '@samples')]

    for i in range(number_of_pcs - 1):
        pc1 = i
        pc2 = i + 1
        plot = figure(
            title='Subpopulation',
            x_axis_label=f'PC{pc1 + 1} ({variance[pc1]})%)',
            y_axis_label=f'PC{pc2 + 1} ({variance[pc2]}%)',
            tooltips=tooltips,
        )
        source = ColumnDataSource(
            dict(
                x=scores.scores[pc1].collect(),
                y=scores.scores[pc2].collect(),
                label=labels,
                samples=sample_names,
            )
        )
        plot.circle(
            'x',
            'y',
            alpha=0.5,
            source=source,
            size=4,
            color=factor_cmap('label', turbo(len(sub_population)), sub_population),
            legend_group='label',
        )
        plot.add_layout(plot.legend[0], 'left')
        plot_filename = out_path_pattern.format(scope='subpop', pci=pc2, ext='png')
        with hl.hadoop_open(plot_filename, 'wb') as f:
            get_screenshot_as_png(plot).save(f, format='PNG')
        html = file_html(plot, CDN, 'my plot')
        plot_filename_html = out_path_pattern.format(
            scope='subpop', pci=pc2, ext='html'
        )
        with hl.hadoop_open(plot_filename_html, 'w') as f:
            f.write(html)

    # Plot loadings
    loadings_ht = hl.read_table(loadings_ht_path)
    for i in range(number_of_pcs - 1):
        pc = i + 1
        plot = manhattan_loadings(
            pvals=hl.abs(loadings_ht.loadings[i]),
            locus=loadings_ht.locus,
            title='Loadings of PC ' + str(pc),
            collect_all=True,
        )
        plot_filename = out_path_pattern.format(scope='loadings', pci=pc, ext='png')
        with hl.hadoop_open(plot_filename, 'wb') as f:
            get_screenshot_as_png(plot).save(f, format='PNG')
        html = file_html(plot, CDN, 'my plot')
        plot_filename_html = out_path_pattern.format(
            scope='loadings', pci=pc, ext='html'
        )
        with hl.hadoop_open(plot_filename_html, 'w') as f:
            f.write(html)


def manhattan_loadings(
    pvals,
    locus=None,
    title=None,
    size=4,
    hover_fields=None,
    collect_all=False,
    n_divisions=500,
):
    """modify hail manhattan plot"""
    palette = [
        '#1f77b4',
        '#ff7f0e',
        '#2ca02c',
        '#d62728',
        '#9467bd',
        '#8c564b',
        '#e377c2',
        '#7f7f7f',
        '#bcbd22',
        '#17becf',
    ]
    if locus is None:
        locus = pvals._indices.source.locus  # pylint: disable=protected-access

    ref = locus.dtype.reference_genome

    if hover_fields is None:
        hover_fields = {}

    hover_fields['locus'] = hl.str(locus)

    source_pd = (
        hl.plot.plots._collect_scatter_plot_data(  # pylint: disable=protected-access
            ('_global_locus', locus.global_position()),
            ('_pval', pvals),
            fields=hover_fields,
            n_divisions=None if collect_all else n_divisions,
        )
    )
    source_pd['p_value'] = source_pd['_pval']
    source_pd['_contig'] = [locus.split(':')[0] for locus in source_pd['locus']]

    observed_contigs = set(source_pd['_contig'])
    observed_contigs = [
        contig for contig in ref.contigs.copy() if contig in observed_contigs
    ]

    contig_ticks = [
        ref._contig_global_position(contig)  # pylint: disable=protected-access
        + ref.contig_length(contig) // 2
        for contig in observed_contigs
    ]
    color_mapper = CategoricalColorMapper(
        factors=ref.contigs, palette=palette[:2] * int((len(ref.contigs) + 1) / 2)
    )

    p = figure(
        title=title, x_axis_label='Chromosome', y_axis_label='Loadings', width=1000
    )
    (
        p,
        _,
        legend,
        _,
        _,
        _,
    ) = hl.plot.plots._get_scatter_plot_elements(  # pylint: disable=protected-access
        p,
        source_pd,
        x_col='_global_locus',
        y_col='_pval',
        label_cols=['_contig'],
        colors={'_contig': color_mapper},
        size=size,
    )
    legend.visible = False
    p.xaxis.ticker = contig_ticks
    p.xaxis.major_label_overrides = dict(zip(contig_ticks, observed_contigs))
    p.select_one(HoverTool).tooltips = [
        t for t in p.select_one(HoverTool).tooltips if not t[0].startswith('_')
    ]

    return p


if __name__ == '__main__':
    main()  # pylint: disable=E1120



#!/usr/bin/env python

"""
Run sample QC on a MatrixTable, hard filter samples, add soft filter labels.
"""

from os.path import join
import logging
from typing import Optional

import click
import hail as hl

from joint_calling import utils
from joint_calling import _version


logger = logging.getLogger(__file__)
logging.basicConfig(format='%(levelname)s (%(name)s %(lineno)s): %(message)s')
logger.setLevel(logging.INFO)


@click.command()
@click.version_option(_version.__version__)
@click.option(
    '--pca-mt',
    'pca_mt_path',
    required=True,
    callback=utils.get_validation_callback(ext='mt', must_exist=True),
    help='path to the Matrix Table ready for PC relate. Dense and GT-annotated',
)
@click.option(
    '--out-relatedness-ht',
    'out_relatedness_ht_path',
)
@click.option(
    '--tmp-bucket',
    'tmp_bucket',
    required=True,
    help='path to write temporary intermediate output and checkpoints',
)
@click.option(
    '--overwrite/--reuse',
    'overwrite',
    is_flag=True,
    help='if an intermediate or a final file exists, skip running the code '
    'that generates it.',
)
@click.option(
    '--hail-billing',
    'hail_billing',
    required=True,
    help='Hail billing account ID.',
)
def main(  # pylint: disable=too-many-arguments,too-many-locals,missing-function-docstring
    pca_mt_path: str,
    out_relatedness_ht_path: str,
    tmp_bucket: str,
    overwrite: bool,
    hail_billing: str,  # pylint: disable=unused-argument
):
    utils.init_hail(__file__)

    mt = hl.read_matrix_table(pca_mt_path)
    mt = mt.select_entries('GT')

    _compute_relatedness(
        mt=mt,
        tmp_bucket=tmp_bucket,
        out_ht_path=out_relatedness_ht_path,
        overwrite=overwrite,
    )


def _compute_relatedness(
    mt: hl.MatrixTable,
    tmp_bucket: str,
    out_ht_path: Optional[str] = None,
    overwrite: bool = False,
) -> hl.Table:
    """
    :param mt: variants selected for PCA analysis (biallelic SNPs, annotated with GT)
    :param tmp_bucket: path to write checkpoints
    :param out_ht_path: path to write relatedness data
    :param overwrite: overwrite checkpoints if they exist
    :return: table with the following structure:
    Row fields:
        'i': str
        'j': str
        'kin': float64
        'ibd0': float64
        'ibd1': float64
        'ibd2': float64
    ----------------------------------------
    Key: ['i', 'j']
    """
    logger.info('Running relatedness check')
    out_ht_path = out_ht_path or join(tmp_bucket, 'relatedness.ht')
    if utils.can_reuse(out_ht_path, overwrite):
        return hl.read_table(out_ht_path)

    scores_ht_path = join(tmp_bucket, 'relatedness_pca_scores.ht')
    if utils.can_reuse(scores_ht_path, overwrite):
        scores_ht = hl.read_table(scores_ht_path)
    else:
        sample_num = mt.cols().count()
        _, scores_ht, _ = hl.hwe_normalized_pca(
            mt.GT, k=max(1, min(sample_num // 3, 10)), compute_loadings=False
        )
        scores_ht.checkpoint(scores_ht_path, overwrite=True)

    relatedness_ht = hl.pc_relate(
        mt.GT,
        min_individual_maf=0.01,
        scores_expr=scores_ht[mt.col_key].scores,
        block_size=4096,
        min_kinship=0.05,
        statistics='all',
    )

    # Converting keys for type struct{str} to str to align
    # with the rank_ht `s` key:
    relatedness_ht = relatedness_ht.key_by(i=relatedness_ht.i.s, j=relatedness_ht.j.s)
    return relatedness_ht.checkpoint(out_ht_path, overwrite=True)


if __name__ == '__main__':
    main()  # pylint: disable=E1120



#!/usr/bin/env python

"""
Creates a Table with all necessary annotations for the random forest model.
Annotations that are included:
    Features for RF:
        - InbreedingCoeff
        - variant_type
        - allele_type
        - n_alt_alleles
        - has_star
        - AS_QD
        - AS_pab_max
        - AS_MQRankSum
        - AS_SOR
        - AS_ReadPosRankSum
    Training sites (bool):
        - transmitted_singleton
        - fail_hard_filters - (ht.QD < 2) | (ht.FS > 60) | (ht.MQ < 30)
"""

from os.path import join
import logging
import click
import hail as hl

from gnomad.resources.grch38.reference_data import get_truth_ht
from gnomad.variant_qc.random_forest import median_impute_features

from joint_calling.utils import get_validation_callback
from joint_calling import utils
from joint_calling import _version

logger = logging.getLogger('random_forest')
logger.setLevel('INFO')


FEATURES = [
    'allele_type',
    'AS_MQRankSum',
    'AS_pab_max',
    'AS_QD',
    'AS_ReadPosRankSum',
    'AS_SOR',
    'InbreedingCoeff',
    'n_alt_alleles',
    'variant_type',
]
INBREEDING_COEFF_HARD_CUTOFF = -0.3
INFO_FEATURES = [
    'AS_MQRankSum',
    'AS_pab_max',
    'AS_QD',
    'AS_ReadPosRankSum',
    'AS_SOR',
]
LABEL_COL = 'rf_label'
PREDICTION_COL = 'rf_prediction'
TRAIN_COL = 'rf_train'
TRUTH_DATA = ['hapmap', 'omni', 'mills', 'kgp_phase1_hc']


@click.command()
@click.version_option(_version.__version__)
@click.option(
    '--info-split-ht',
    'info_split_ht_path',
    required=True,
    callback=get_validation_callback(ext='ht', must_exist=True),
    help='path to info Table with split multiallelics '
    '(generated by generate_variant_qc_annotations.py --split-multiallelic)',
)
@click.option(
    '--freq-ht',
    'freq_ht_path',
    required=True,
    callback=get_validation_callback(ext='ht', must_exist=True),
    help='path to a Table with InbreedingCoeff (generated by generate_freq_data.py)',
)
@click.option(
    '--fam-stats-ht',
    'fam_stats_ht_path',
    callback=get_validation_callback(ext='ht'),
    help='optional path to a Table with trio stats '
    '(generated by generate_variant_qc_annotations.py)',
)
@click.option(
    '--allele-data-ht',
    'allele_data_ht_path',
    required=True,
    callback=get_validation_callback(ext='ht', must_exist=True),
    help='path to a Table with allele data '
    '(generated by generate_variant_qc_annotations.py)',
)
@click.option(
    '--qc-ac-ht',
    'qc_ac_ht_path',
    required=True,
    callback=get_validation_callback(ext='ht', must_exist=True),
    help='path to a Table with allele counts '
    '(generated by generate_variant_qc_annotations.py)',
)
@click.option(
    '--out-ht',
    'out_ht_path',
    required=True,
    callback=get_validation_callback(ext='ht'),
    help='Path to write the result',
)
@click.option(
    '--bucket',
    'work_bucket',
    required=True,
    help='path to write intermediate output and checkpoints. '
    'Can be a Google Storage URL (i.e. start with `gs://`).',
)
@click.option(
    '--local-tmp-dir',
    'local_tmp_dir',
    help='local directory for temporary files and Hail logs (must be local).',
)
@click.option(
    '--impute-features',
    'impute_features',
    is_flag=True,
    help='If set, feature imputation is performed',
)
@click.option(
    '--n-partitions',
    'n_partitions',
    type=click.INT,
    help='Desired base number of partitions for output tables',
    default=5000,
)
@click.option(
    '--use-adj-genotypes',
    'use_adj_genotypes',
    help='Use adj genotypes',
    is_flag=True,
)
@click.option(
    '--overwrite/--reuse',
    'overwrite',
    is_flag=True,
    help='if an intermediate or a final file exists, skip running the code '
    'that generates it.',
)
def main(  # pylint: disable=too-many-arguments,too-many-locals
    info_split_ht_path: str,
    freq_ht_path: str,
    fam_stats_ht_path: str,
    allele_data_ht_path: str,
    qc_ac_ht_path: str,
    out_ht_path: str,
    work_bucket: str,
    local_tmp_dir: str,
    impute_features: bool,
    n_partitions: int,
    use_adj_genotypes: bool,
    overwrite: bool,  # pylint: disable=unused-argument
):  # pylint: disable=missing-function-docstring
    local_tmp_dir = utils.init_hail('variant_qc_random_forest', local_tmp_dir)

    group = 'adj' if use_adj_genotypes else 'raw'
    ht = hl.read_table(info_split_ht_path)
    ht = ht.transmute(**ht.info)
    ht = ht.select('lowqual', 'AS_lowqual', 'FS', 'MQ', 'QD', *INFO_FEATURES)

    inbreeding_ht = hl.read_table(freq_ht_path)
    inbreeding_ht = inbreeding_ht.select(
        InbreedingCoeff=hl.if_else(
            hl.is_nan(inbreeding_ht.InbreedingCoeff),
            hl.null(hl.tfloat32),
            inbreeding_ht.InbreedingCoeff,
        )
    )
    if fam_stats_ht_path:
        fam_stats_ht = hl.read_table(fam_stats_ht_path)
        fam_stats_ht = fam_stats_ht.select(
            f'n_transmitted_{group}', f'ac_children_{group}'
        )
        ht = ht.annotate(
            **fam_stats_ht[ht.key],
        )

    logger.info('Annotating Table with all columns from multiple annotation Tables')
    truth_data_ht = get_truth_ht()
    allele_data_ht = hl.read_table(allele_data_ht_path)
    qc_ac_ht = hl.read_table(qc_ac_ht_path)
    ht = ht.annotate(
        **inbreeding_ht[ht.key],
        **truth_data_ht[ht.key],
        **allele_data_ht[ht.key].allele_data,
        **qc_ac_ht[ht.key],
    )

    # Filter to only variants found in high quality samples and are not lowqual
    ht = ht.filter((ht[f'ac_qc_samples_{group}'] > 0) & ~ht.AS_lowqual)
    ht = ht.select(
        'a_index',
        'was_split',
        *FEATURES,
        *TRUTH_DATA,
        fail_hard_filters=(ht.QD < 2) | (ht.FS > 60) | (ht.MQ < 30),
        transmitted_singleton=(
            (ht[f'n_transmitted_{group}'] == 1) & (ht[f'ac_qc_samples_{group}'] == 2)
        )
        if fam_stats_ht_path is not None
        else hl.literal(False),
        singleton=(ht.ac_release_samples_raw == 1),
        ac_raw=ht.ac_qc_samples_raw,
        ac=ht.ac_release_samples_adj,
        ac_qc_samples_unrelated_raw=ht.ac_qc_samples_unrelated_raw,
    )

    ht = ht.repartition(n_partitions, shuffle=False)
    if work_bucket:
        checkpoint_path = join(work_bucket, 'rf-annotations-before-impute.ht')
        ht = ht.checkpoint(checkpoint_path, overwrite=True)

    if impute_features:
        ht = median_impute_features(ht, {'variant_type': ht.variant_type})

    ht.write(out_ht_path, overwrite=True)

    summary = ht.group_by(
        'omni',
        'mills',
        'transmitted_singleton',
    ).aggregate(n=hl.agg.count())
    logger.info('Summary of truth data annotations:')
    summary.show(20)


if __name__ == '__main__':
    main()  # pylint: disable=E1120



#!/usr/bin/env python

"""
Apply RF-based filters to a matrix table. Must be run after evaluation.py
"""

import logging
from typing import Optional

import hail as hl
import click

from gnomad.utils.filtering import add_filters_expr
from gnomad.variant_qc.pipeline import INBREEDING_COEFF_HARD_CUTOFF

from joint_calling.utils import get_validation_callback
from joint_calling import utils
from joint_calling import _version


logger = logging.getLogger('variant_qc_filtering')
logger.setLevel('INFO')


# RF or VQSR score bin to use as cutoff for SNPs. Value should be between 1 and 100
SNP_BIN_CUTOFF = 90
# RF or VQSR score to use as cutoff for SNPs. Mutually exclusive to snp_bin_cutoff
SNP_SCORE_CUTOFF = None
# RF or VQSR score bin to use as cutoff for indels. Value should be between 1 and 100
INDEL_BIN_CUTOFF = 80
# RF or VQSR score to use as cutoff for indels. Mutually exclusive to indel_bin_cutoff
INDEL_SCORE_CUTOFF = None
# Filter centromeres and telomeres from final filter Table
FILTER_CENTROMERE_TELOMERE = True


@click.command()
@click.version_option(_version.__version__)
@click.option(
    '--out-final-filter-ht',
    'out_final_filter_ht_path',
    required=True,
    callback=get_validation_callback(ext='ht'),
    help='Path to write the final filter HT to',
)
@click.option('--model-id', 'model_id', help='Filtering model ID to use.')
@click.option(
    '--model-name',
    'model_name',
    help='Filtering model name to use in the filters field. Either "VQSR" or "RF"',
    type=click.Choice(['VQSR', 'RF']),
)
@click.option(
    '--score-name',
    'score_name',
    help=(
        'What to rename the filtering score annotation. This will be used in place '
        'of "score" in the release HT info struct and the INFO field of the VCF '
        '(e.g. "RF", "AS_VQSLOD").'
    ),
)
@click.option(
    '--info-split-ht',
    'info_split_ht_path',
    required=True,
    callback=get_validation_callback(ext='ht', must_exist=True),
    help='path to info Table with split multiallelics '
    '(generated by generate_variant_qc_annotations --split-multiallelic)',
)
@click.option(
    '--freq-ht',
    'freq_ht_path',
    required=True,
    callback=get_validation_callback(ext='ht', must_exist=True),
    help='path to a Table with InbreedingCoeff (generated by generate_freq_data.py)',
)
@click.option(
    '--score-bin-ht',
    'score_bin_ht_path',
    callback=get_validation_callback(ext='ht', must_exist=True),
    required=True,
    help='Generated by evaluation.py',
)
@click.option(
    '--score-bin-agg-ht',
    'score_bin_agg_ht_path',
    callback=get_validation_callback(ext='ht', must_exist=True),
    required=True,
    help='Generated by evaluation.py',
)
@click.option(
    '--vqsr-filters-split-ht',
    'vqsr_filters_split_ht_path',
    callback=get_validation_callback(ext='ht'),
    help='Use VQSR training sites to train the RF (generated by load_vqsr.py)'
    'A "vqsr" annotation will be added to the '
    'final filter Table containing AS_VQSLOD, AS_culprit, NEGATIVE_TRAIN_SITE '
    'and POSITIVE_TRAIN_SITE.',
)
@click.option(
    '--inbreeding-coeff-threshold',
    'inbreeding_coeff_threshold',
    help='InbreedingCoeff hard filter to use for variants.',
    type=float,
    default=INBREEDING_COEFF_HARD_CUTOFF,
)
# @click.option(  # NOTE: This was required for v3.1 to grab the SOR annotation, we now compute this in `get_site_info_expr`
#     '--vqsr-model-id',
#     'vqsr_model_id',
#     help=(
#         'If a VQSR model ID is provided, a 'vqsr' annotation will be added to the '
#         'final filter Table containing AS_VQSLOD, AS_culprit, NEGATIVE_TRAIN_SITE '
#         'and POSITIVE_TRAIN_SITE.'
#     ),
#     default='vqsr_alleleSpecificTrans',
#     type=click.Choice(['vqsr_classic', 'vqsr_alleleSpecific', 'vqsr_alleleSpecificTrans']),
# )
@click.option(
    '--bucket',
    'work_bucket',
    required=True,
    help='path to write intermediate output and checkpoints',
)
@click.option(
    '--local-tmp-dir',
    'local_tmp_dir',
    help='local directory for temporary files and Hail logs (must be local).',
)
@click.option(
    '--overwrite/--reuse',
    'overwrite',
    is_flag=True,
    help='if an intermediate or a final file exists, skip running the code '
    'that generates it.',
)
def main(
    out_final_filter_ht_path,
    model_id,
    model_name,
    score_name,
    info_split_ht_path,
    freq_ht_path,
    score_bin_ht_path,
    score_bin_agg_ht_path,
    vqsr_filters_split_ht_path,
    inbreeding_coeff_threshold,
    work_bucket: str,  # pylint: disable=unused-argument
    local_tmp_dir: str,
    overwrite,
):  # pylint: disable=missing-function-docstring
    local_tmp_dir = utils.init_hail('variant_qc_finalize', local_tmp_dir)

    if not overwrite and utils.file_exists(out_final_filter_ht_path):
        ht = hl.read_table(out_final_filter_ht_path)
        ht.summarize()
    else:
        ht = hl.read_table(score_bin_ht_path)
        if FILTER_CENTROMERE_TELOMERE:
            tel_cent_ht = hl.read_table(utils.TEL_AND_CENT_HT_PATH)
            ht = ht.filter(~hl.is_defined(tel_cent_ht[ht.locus]))

        info_ht = hl.read_table(info_split_ht_path)
        ht = ht.filter(~info_ht[ht.key].AS_lowqual)

        if model_id.startswith('vqsr_'):
            ht = ht.drop('info')

        freq_ht = hl.read_table(freq_ht_path)
        ht = ht.annotate(InbreedingCoeff=freq_ht[ht.key].InbreedingCoeff)
        freq_idx = freq_ht[ht.key]
        aggregated_bin_ht = hl.read_table(score_bin_agg_ht_path)

        vqsr_ht = None
        if vqsr_filters_split_ht_path:
            vqsr_ht = hl.read_table(vqsr_filters_split_ht_path)

        ht = generate_final_filter_ht(
            ht,
            model_name,
            score_name,
            ac0_filter_expr=freq_idx.freq[0].AC == 0,
            ts_ac_filter_expr=freq_idx.freq[1].AC == 1,
            mono_allelic_flag_expr=(freq_idx.freq[1].AF == 1)
            | (freq_idx.freq[1].AF == 0),
            snp_bin_cutoff=SNP_BIN_CUTOFF,
            indel_bin_cutoff=INDEL_BIN_CUTOFF,
            snp_score_cutoff=SNP_SCORE_CUTOFF,
            indel_score_cutoff=INDEL_SCORE_CUTOFF,
            inbreeding_coeff_cutoff=inbreeding_coeff_threshold,
            aggregated_bin_ht=aggregated_bin_ht,
            bin_id='bin',
            vqsr_ht=vqsr_ht,
        )
        ht = ht.annotate_globals(
            filtering_model=ht.filtering_model.annotate(model_id=model_id)
        )
        if model_id.startswith('vqsr_'):
            ht = ht.annotate_globals(
                filtering_model=ht.filtering_model.annotate(
                    snv_training_variables=[
                        'AS_QD',
                        'AS_MQRankSum',
                        'AS_ReadPosRankSum',
                        'AS_FS',
                        'AS_SOR',
                        'AS_MQ',
                    ],
                    indel_training_variables=[
                        'AS_QD',
                        'AS_MQRankSum',
                        'AS_ReadPosRankSum',
                        'AS_FS',
                        'AS_SOR',
                    ],
                )
            )
        else:
            ht = ht.annotate_globals(
                filtering_model=ht.filtering_model.annotate(
                    snv_training_variables=ht.features,
                    indel_training_variables=ht.features,
                )
            )

        ht = ht.checkpoint(out_final_filter_ht_path, True)
        ht.summarize()


def generate_final_filter_ht(
    ht: hl.Table,
    model_name: str,
    score_name: str,
    ac0_filter_expr: hl.expr.BooleanExpression,
    ts_ac_filter_expr: hl.expr.BooleanExpression,
    mono_allelic_flag_expr: hl.expr.BooleanExpression,
    inbreeding_coeff_cutoff: float = INBREEDING_COEFF_HARD_CUTOFF,
    snp_bin_cutoff: int = None,
    indel_bin_cutoff: int = None,
    snp_score_cutoff: float = None,
    indel_score_cutoff: float = None,
    aggregated_bin_ht: Optional[hl.Table] = None,
    bin_id: Optional[str] = None,
    vqsr_ht: hl.Table = None,
) -> hl.Table:
    """
    Prepares finalized filtering model given a filtering HT from `rf.apply_rf_model`
    or VQSR and cutoffs for filtering.

    .. note::

        - `snp_bin_cutoff` and `snp_score_cutoff` are mutually exclusive,
          and one must be supplied.
        - `indel_bin_cutoff` and `indel_score_cutoff` are mutually exclusive,
           and one must be supplied.
        - If a `snp_bin_cutoff` or `indel_bin_cutoff` cutoff is supplied then an
          `aggregated_bin_ht` and `bin_id` must also be supplied to determine the SNP
          and indel scores to use as cutoffs from an aggregated bin Table like
          one created by `compute_grouped_binned_ht` in combination with
          `score_bin_agg`.

    :param ht: Filtering Table from `rf.apply_rf_model` or VQSR to prepare as the final
           filter Table
    :param model_name: Filtering model name to use in the 'filters' field (VQSR or RF)
    :param score_name: Name to use for the filtering score annotation. This will be
           used in place of 'score' in the release HT info struct and the INFO field
           of the VCF (e.g. RF or AS_VQSLOD)
    :param ac0_filter_expr: Expression that indicates if a variant should be filtered
           as allele count 0 (AC0)
    :param ts_ac_filter_expr: Allele count expression in `ht` to use as a filter for
           determining a transmitted singleton
    :param mono_allelic_flag_expr: Expression indicating if a variant is mono-allelic
    :param inbreeding_coeff_cutoff: InbreedingCoeff hard filter to use for variants
    :param snp_bin_cutoff: Bin cutoff to use for SNP variant QC filter. Can't be used
           with `snp_score_cutoff`
    :param indel_bin_cutoff: Bin cutoff to use for indel variant QC filter. Can't be
           used with `indel_score_cutoff`
    :param snp_score_cutoff: Score cutoff (e.g. RF probability or AS_VQSLOD) to use
           for SNP variant QC filter. Can't be used with `snp_bin_cutoff`
    :param indel_score_cutoff: Score cutoff (e.g. RF probability or AS_VQSLOD) to use
           for indel variant QC filter. Can't be used with `indel_bin_cutoff`
    :param aggregated_bin_ht: Table with aggregate counts of variants based on bins
    :param bin_id: Name of bin to use in 'bin_id' column of `aggregated_bin_ht` to use
           to determine probability cutoff
    :param vqsr_ht: If a VQSR HT is supplied a 'vqsr' annotation containing AS_VQSLOD,
           AS_culprit, NEGATIVE_TRAIN_SITE, and POSITIVE_TRAIN_SITE will be included
           in the returned Table
    :return: Finalized random forest Table annotated with variant filters
    """
    if snp_bin_cutoff is not None and snp_score_cutoff is not None:
        raise ValueError(
            'snp_bin_cutoff and snp_score_cutoff are mutually exclusive, please only supply one SNP filtering cutoff.'
        )

    if indel_bin_cutoff is not None and indel_score_cutoff is not None:
        raise ValueError(
            'indel_bin_cutoff and indel_score_cutoff are mutually exclusive, please only supply one indel filtering cutoff.'
        )

    if snp_bin_cutoff is None and snp_score_cutoff is None:
        raise ValueError(
            'One (and only one) of the parameters snp_bin_cutoff and snp_score_cutoff must be supplied.'
        )

    if indel_bin_cutoff is None and indel_score_cutoff is None:
        raise ValueError(
            'One (and only one) of the parameters indel_bin_cutoff and indel_score_cutoff must be supplied.'
        )

    if (snp_bin_cutoff is not None or indel_bin_cutoff is not None) and (
        aggregated_bin_ht is None or bin_id is None
    ):
        raise ValueError(
            'If using snp_bin_cutoff or indel_bin_cutoff, both aggregated_bin_ht and bin_id must be supplied'
        )

    min_score = ht.aggregate(hl.agg.min(ht.score))
    max_score = ht.aggregate(hl.agg.max(ht.score))

    # Determine SNP and indel score cutoffs if given bin instead of score
    if snp_bin_cutoff:
        assert aggregated_bin_ht
        snp_score_cutoff = aggregated_bin_ht.aggregate(
            hl.agg.filter(
                aggregated_bin_ht.snv
                & (aggregated_bin_ht.bin_id == bin_id)
                & (aggregated_bin_ht.bin == snp_bin_cutoff),
                hl.agg.min(aggregated_bin_ht.min_score),
            )
        )
        snp_cutoff_global = hl.struct(bin=snp_bin_cutoff, min_score=snp_score_cutoff)
    elif snp_score_cutoff:
        if snp_score_cutoff < min_score or snp_score_cutoff > max_score:
            raise ValueError('snp_score_cutoff is not within the range of score.')
        snp_cutoff_global = hl.struct(min_score=snp_score_cutoff)
    else:
        raise ValueError('Either snp_bin_cutoff or snp_score_cutoff must be set')

    if indel_bin_cutoff:
        assert aggregated_bin_ht
        indel_score_cutoff = aggregated_bin_ht.aggregate(
            hl.agg.filter(
                ~aggregated_bin_ht.snv
                & (aggregated_bin_ht.bin_id == bin_id)
                & (aggregated_bin_ht.bin == indel_bin_cutoff),
                hl.agg.min(aggregated_bin_ht.min_score),
            )
        )
        indel_cutoff_global = hl.struct(
            bin=indel_bin_cutoff, min_score=indel_score_cutoff
        )
    elif indel_score_cutoff:
        if indel_score_cutoff < min_score or indel_score_cutoff > max_score:
            raise ValueError('indel_score_cutoff is not within the range of score.')
        indel_cutoff_global = hl.struct(min_score=indel_score_cutoff)
    else:
        raise ValueError('Either indel_bin_cutoff or indel_score_cutoff must be set')

    logger.info(
        f'Using a SNP score cutoff of {snp_score_cutoff} and an indel score cutoff of {indel_score_cutoff}.'
    )

    # Add filters to HT
    filters = dict()

    if ht.any(hl.is_missing(ht.score)):
        ht.filter(hl.is_missing(ht.score)).show()
        raise ValueError('Missing Score!')

    filters[model_name] = (
        hl.is_missing(ht.score)
        | (
            hl.is_snp(ht.alleles[0], ht.alleles[1])
            & (ht.score < snp_cutoff_global.min_score)
        )
        | (
            ~hl.is_snp(ht.alleles[0], ht.alleles[1])
            & (ht.score < indel_cutoff_global.min_score)
        )
    )

    filters['InbreedingCoeff'] = hl.or_else(
        ht.InbreedingCoeff < inbreeding_coeff_cutoff, False
    )
    filters['AC0'] = ac0_filter_expr

    annotations_expr = dict()
    if model_name == 'RF':
        # Fix annotations for release
        annotations_expr.update(
            {
                'positive_train_site': hl.or_else(ht.positive_train_site, False),
                'rf_tp_probability': ht.rf_probability['TP'],
            }
        )
    annotations_expr.update(
        {
            'transmitted_singleton': hl.or_missing(
                ts_ac_filter_expr, ht.transmitted_singleton
            )
        }
    )
    if 'feature_imputed' in ht.row:
        annotations_expr.update(
            {
                x: hl.or_missing(~ht.feature_imputed[x], ht[x])
                for x in ht.row.feature_imputed
            }
        )

    ht = ht.transmute(
        filters=add_filters_expr(filters=filters),
        monoallelic=mono_allelic_flag_expr,
        **{score_name: ht.score},
        **annotations_expr,
    )

    bin_names = [x for x in ht.row if x.endswith('bin')]
    bin_names = [
        (
            x,
            x.split('adj_')[0] + x.split('adj_')[1]
            if len(x.split('adj_')) == 2
            else 'raw_' + x,
        )
        for x in bin_names
    ]
    ht = ht.transmute(**{j: ht[i] for i, j in bin_names})

    ht = ht.annotate_globals(
        bin_group_variant_counts=hl.struct(
            **{j: ht.bin_group_variant_counts[i] for i, j in bin_names}
        ),
        filtering_model=hl.struct(
            model_name=model_name,
            score_name=score_name,
            snv_cutoff=snp_cutoff_global,
            indel_cutoff=indel_cutoff_global,
        ),
        inbreeding_coeff_cutoff=inbreeding_coeff_cutoff,
    )
    if vqsr_ht:
        vqsr = vqsr_ht[ht.key]
        ht = ht.annotate(
            vqsr=hl.struct(
                AS_VQSLOD=vqsr.info.AS_VQSLOD,
                AS_culprit=vqsr.info.AS_culprit,
                NEGATIVE_TRAIN_SITE=vqsr.info.NEGATIVE_TRAIN_SITE,
                POSITIVE_TRAIN_SITE=vqsr.info.POSITIVE_TRAIN_SITE,
            ),
            SOR=vqsr.info.SOR,  # NOTE: This was required for v3.1,
            # we now compute this in `get_site_info_expr`
        )
        ht = ht.drop('AS_culprit')

    return ht


if __name__ == '__main__':
    main()  # pylint: disable=E1120



#!/usr/bin/env python

"""
Run Random Forest variant QC
"""

from os.path import join
from typing import List, Optional, Tuple
import json
import logging
import uuid
import pyspark
import click
import hail as hl

from gnomad.variant_qc.pipeline import train_rf_model
from gnomad.variant_qc.random_forest import (
    apply_rf_model,
    get_rf_runs,
    get_run_data,
    load_model,
    pretty_print_runs,
    save_model,
)

from joint_calling.utils import file_exists, get_validation_callback
from joint_calling import utils
from joint_calling import _version

logger = logging.getLogger('random_forest')
logger.setLevel('INFO')


FEATURES = [
    'allele_type',
    'AS_MQRankSum',
    'AS_pab_max',
    'AS_QD',
    'AS_ReadPosRankSum',
    'AS_SOR',
    'InbreedingCoeff',
    'n_alt_alleles',
    'variant_type',
]
INBREEDING_COEFF_HARD_CUTOFF = -0.3
INFO_FEATURES = [
    'AS_MQRankSum',
    'AS_pab_max',
    'AS_QD',
    'AS_ReadPosRankSum',
    'AS_SOR',
]
LABEL_COL = 'rf_label'
PREDICTION_COL = 'rf_prediction'
TRAIN_COL = 'rf_train'
TRUTH_DATA = ['hapmap', 'omni', 'mills', 'kgp_phase1_hc']


@click.command()
@click.version_option(_version.__version__)
@click.option(
    '--annotations-ht',
    'annotations_ht_path',
    required=True,
    callback=get_validation_callback(ext='ht', must_exist=True),
    help='RF annotations, generated by create_rf_annotations.py',
)
@click.option(
    '--vqsr-filters-split-ht',
    'vqsr_filters_split_ht_path',
    callback=get_validation_callback(ext='ht'),
    help='Use VQSR training sites to train the RF (generated by load_vqsr.py)',
)
@click.option(
    '--out-results-ht',
    'out_results_ht_path',
    required=True,
    callback=get_validation_callback(ext='ht'),
    help='Path to write to RF results to',
)
@click.option(
    '--out-model-id',
    'model_id',
    help='Model ID for the output models',
)
@click.option(
    '--bucket',
    'work_bucket',
    required=True,
    help='path to write intermediate output and checkpoints. '
    'Can be a Google Storage URL (i.e. start with `gs://`).',
)
@click.option(
    '--local-tmp-dir',
    'local_tmp_dir',
    help='local directory for temporary files and Hail logs (must be local).',
)
@click.option(
    '--overwrite/--reuse',
    'overwrite',
    is_flag=True,
    help='if an intermediate or a final file exists, skip running the code '
    'that generates it.',
)
@click.option(
    '--model-file',
    'model_file_path',
    help='Pre-created model file (requires --model-id and --model-training-ht)',
)
@click.option(
    '--model-training-ht',
    'model_training_ht_path',
    help='Pre-created model training file (requires --model-id and --model-file)',
)
@click.option(
    '--fp-to-tp',
    'fp_to_tp',
    help='Ratio of FPs to TPs for training the RF model. If 0, '
    'all training examples are used. (default=1.0)',
    type=click.FLOAT,
    default=1.0,
)
@click.option(
    '--test-intervals',
    'test_intervals',
    help='The specified interval(s) will be held out for testing and '
    'evaluation only. (default to "chr20")',
    multiple=True,
    type=click.STRING,
    default=['chr20'],
)
@click.option(
    '--num-trees',
    'num_trees',
    type=click.INT,
    help='Number of trees in the RF model',
    default=500,
)
@click.option(
    '--max-depth',
    'max_depth',
    help='Maxmimum tree depth in the RF model',
    default=5,
    type=click.INT,
)
# Training data parameters
@click.option(
    '--use-adj-genotypes',
    'use_adj_genotypes',
    help='Use adj genotypes',
    is_flag=True,
)
@click.option(
    '--vqsr-model-id',
    'vqsr_model_id',
    help='If a VQSR model ID is provided the VQSR training annotations will be '
    'used for training.',
    default='vqsr_alleleSpecificTrans',
    type=click.Choice(
        ['vqsr_classic', 'vqsr_alleleSpecific', 'vqsr_alleleSpecificTrans'],
        case_sensitive=False,
    ),
)
@click.option(
    '--no-transmitted-singletons',
    'no_transmitted_singletons',
    help='Do not use transmitted singletons for training.',
    is_flag=True,
)
@click.option(
    '--no-inbreeding-coeff',
    'no_inbreeding_coeff',
    help='Train RF without inbreeding coefficient as a feature.',
    is_flag=True,
)
@click.option(
    '--filter-centromere-telomere',
    'filter_centromere_telomere',
    help='Train RF without centromeres and telomeres.',
    is_flag=True,
)
def main(  # pylint: disable=too-many-arguments,too-many-locals,missing-function-docstring
    annotations_ht_path: str,
    vqsr_filters_split_ht_path: Optional[str],
    out_results_ht_path: str,
    model_id: str,
    work_bucket: str,
    local_tmp_dir: str,
    overwrite: bool,
    model_file_path: Optional[str],
    model_training_ht_path: Optional[str],
    fp_to_tp: Optional[float],
    test_intervals: Optional[List[str]],
    num_trees: Optional[int],
    max_depth: Optional[int],
    use_adj_genotypes: bool,
    vqsr_model_id: Optional[str],  # pylint: disable=unused-argument
    no_transmitted_singletons: Optional[bool],
    no_inbreeding_coeff: Optional[bool],
    filter_centromere_telomere: Optional[bool],
):
    local_tmp_dir = utils.init_hail('variant_qc_random_forest', local_tmp_dir)

    rf_json_path = join(work_bucket, 'rf_runs.json')
    rf_runs = get_rf_runs(rf_json_path)
    if rf_runs:
        logger.info(f'RF runs:')
        pretty_print_runs(rf_runs)

    if model_file_path:
        # Overwriting, but want to reuse preexisting model
        assert model_file_path and file_exists(model_file_path)
        assert model_training_ht_path and file_exists(model_training_ht_path)
        logger.info(
            f'Loading provided model file {model_file_path}, '
            f'training ht {model_training_ht_path}'
        )
        model = load_model(model_file_path)
        training_ht = hl.read_table(model_training_ht_path)
    else:
        # Train a new model
        logger.info('Training new model')
        model, model_id, training_ht = train_model(
            annotations_ht=hl.read_table(annotations_ht_path),
            rf_json_path=rf_json_path,
            model_id=model_id,
            work_bucket=work_bucket,
            overwrite=overwrite,
            use_adj_genotypes=use_adj_genotypes,
            filter_centromere_telomere=filter_centromere_telomere,
            no_transmitted_singletons=no_transmitted_singletons,
            no_inbreeding_coeff=no_inbreeding_coeff,
            fp_to_tp=fp_to_tp,
            max_depth=max_depth,
            num_trees=num_trees,
            test_intervals=test_intervals,
            vqsr_filters_split_ht_path=vqsr_filters_split_ht_path,
        )
        logger.info('Done training model')

    logger.info(f'Applying RF model...')
    ht = apply_rf_model(
        training_ht,
        rf_model=model,
        features=hl.eval(training_ht.features),
        label=LABEL_COL,
    )

    ht.write(out_results_ht_path, overwrite=True)

    summary_ht = ht.group_by(
        'tp', 'fp', TRAIN_COL, LABEL_COL, PREDICTION_COL
    ).aggregate(n=hl.agg.count())

    summary_ht.show(n=20)


def train_model(
    annotations_ht: hl.Table,
    rf_json_path: str,
    model_id: Optional[str],
    work_bucket: str,
    overwrite: bool,
    use_adj_genotypes: bool,
    filter_centromere_telomere: Optional[bool],
    no_transmitted_singletons: Optional[bool],
    no_inbreeding_coeff: Optional[bool],
    fp_to_tp: Optional[float],
    max_depth: Optional[int],
    num_trees: Optional[int],
    test_intervals: Optional[List[str]],
    vqsr_filters_split_ht_path: Optional[str],
) -> Tuple[hl.Table, str, pyspark.ml.PipelineModel]:
    """
    Train RF model
    :param annotations_ht:
    :param rf_json_path:
    :param model_id:
    :param work_bucket:
    :param overwrite:
    :param use_adj_genotypes:
    :param filter_centromere_telomere:
    :param no_transmitted_singletons:
    :param no_inbreeding_coeff:
    :param fp_to_tp:
    :param max_depth:
    :param num_trees:
    :param test_intervals:
    :param vqsr_filters_split_ht_path:
    :return: model, model_id, training_ht
    """
    rf_runs = get_rf_runs(rf_json_path)
    if not model_id:
        model_id = f'rf_{str(uuid.uuid4())[:8]}'
        while model_id in rf_runs:
            model_id = f'rf_{str(uuid.uuid4())[:8]}'
    training_ht, model = train_rf(
        annotations_ht,
        vqsr_filters_split_ht=hl.read_table(vqsr_filters_split_ht_path)
        if vqsr_filters_split_ht_path
        else None,
        fp_to_tp=fp_to_tp,
        num_trees=num_trees,
        max_depth=max_depth,
        no_transmitted_singletons=no_transmitted_singletons,
        no_inbreeding_coeff=no_inbreeding_coeff,
        filter_centromere_telomere=filter_centromere_telomere,
        test_intervals=test_intervals,
    )
    out_model_training_ht_path = join(
        work_bucket, 'models', f'{model_id}-rf_training.ht'
    )
    out_model_file_path = join(work_bucket, 'models', f'{model_id}-rf.model')

    logger.info(f'Saving training Table to {out_model_training_ht_path}')
    training_ht = training_ht.checkpoint(
        out_model_training_ht_path, overwrite=overwrite
    )
    logger.info('Adding run to RF run list')
    rf_runs[model_id] = get_run_data(
        input_args={
            'transmitted_singletons': None
            if vqsr_filters_split_ht_path
            else not no_transmitted_singletons,
            'adj': use_adj_genotypes,
            'vqsr_training': vqsr_filters_split_ht_path is not None,
            'filter_centromere_telomere': filter_centromere_telomere,
        },
        test_intervals=test_intervals if test_intervals else None,
        features_importance=dict(hl.eval(training_ht.features_importance)),
        test_results=hl.eval(training_ht.test_results),
    )
    with hl.hadoop_open(rf_json_path, 'w') as f:
        json.dump(rf_runs, f)

    logger.info(f'Saving RF model to {out_model_file_path}')
    save_model(model, out_model_file_path, overwrite=overwrite)
    return model, model_id, training_ht


def train_rf(
    ht: hl.Table,
    vqsr_filters_split_ht: Optional[hl.Table],
    fp_to_tp: Optional[float] = 1.0,
    num_trees: Optional[int] = 500,
    max_depth: Optional[int] = 5,
    no_transmitted_singletons: Optional[bool] = False,
    no_inbreeding_coeff: Optional[bool] = False,
    filter_centromere_telomere: Optional[bool] = False,
    test_intervals: Optional[List[str]] = None,
) -> Tuple[hl.Table, pyspark.ml.PipelineModel]:
    """
    Train random forest model using `train_rf_model`
    :param ht: Table containing annotations needed for RF training, built with `create_rf_ht()`
    :param vqsr_filters_split_ht: Use VQSR these training sites to train the RF. Table is build with load_vqsr.py
    :param fp_to_tp: Ratio of FPs to TPs for creating the RF model. If set to 0, all training examples are used.
    :param num_trees: Number of trees in the RF model.
    :param max_depth: Maxmimum tree depth in the RF model.
    :param no_transmitted_singletons: Do not use transmitted singletons for training.
    :param no_inbreeding_coeff: Do not use inbreeding coefficient as a feature for training.
    :param filter_centromere_telomere: Filter centromeres and telomeres before training.
    :param test_intervals: Specified interval(s) will be held out for testing and evaluation only. (default to 'chr20')
    :return: `ht` annotated with training information and the RF model
    """
    features = FEATURES
    if no_inbreeding_coeff:
        logger.info('Removing InbreedingCoeff from list of features...')
        features.remove('InbreedingCoeff')

    if vqsr_filters_split_ht is not None:
        logger.info('Using VQSR training sites for RF training...')
        ht = ht.annotate(
            vqsr_POSITIVE_TRAIN_SITE=vqsr_filters_split_ht[
                ht.key
            ].info.POSITIVE_TRAIN_SITE,
            vqsr_NEGATIVE_TRAIN_SITE=vqsr_filters_split_ht[
                ht.key
            ].info.NEGATIVE_TRAIN_SITE,
        )
        tp_expr = ht.vqsr_POSITIVE_TRAIN_SITE
        fp_expr = ht.vqsr_NEGATIVE_TRAIN_SITE
    else:
        fp_expr = ht.fail_hard_filters
        tp_expr = ht.omni | ht.mills
        if not no_transmitted_singletons:
            tp_expr = tp_expr | ht.transmitted_singleton

    if test_intervals is None:
        test_intervals = ['chr20']
    logger.info(f'Using test intervals: {test_intervals}')
    test_intervals = [
        hl.parse_locus_interval(x, reference_genome='GRCh38') for x in test_intervals
    ]

    ht = ht.annotate(tp=tp_expr, fp=fp_expr)

    if filter_centromere_telomere:
        logger.info('Filtering centromeres and telomeres from HT...')
        rf_ht = ht.filter(
            ~hl.is_defined(hl.read_table(utils.TEL_AND_CENT_HT_PATH)[ht.locus])
        )
    else:
        rf_ht = ht

    trained_rf_ht, rf_model = train_rf_model(
        rf_ht,
        rf_features=features,
        tp_expr=rf_ht.tp,
        fp_expr=rf_ht.fp,
        fp_to_tp=fp_to_tp,
        num_trees=num_trees,
        max_depth=max_depth,
        test_expr=hl.literal(test_intervals).any(
            lambda interval: interval.contains(rf_ht.locus)
        ),
    )
    logger.info('Joining original RF Table with training information')
    ht = ht.join(trained_rf_ht, how='left')
    return ht, rf_model


if __name__ == '__main__':
    main()  # pylint: disable=E1120



"""
Dataproc cluster utils
"""
from typing import Optional, List
from analysis_runner import dataproc
from hailtop.batch.job import Job
from hailtop.batch import Batch
from joint_calling import utils


clusters_by_name = dict()


def get_cluster(
    b: Batch,
    name,
    scatter_count: int,
    long: bool = False,
    is_test: bool = False,
    phantomjs: bool = False,
    preempt: bool = False,
    depends_on: Optional[List[Job]] = None,
):
    """
    Get or create a Dataproc cluster by name
    """
    if name not in clusters_by_name:
        max_age = '1h' if is_test else '12h'
        if long:
            max_age = '2h' if is_test else '24h'

        clusters_by_name[name] = dataproc.setup_dataproc(
            b,
            max_age=max_age,
            packages=utils.DATAPROC_PACKAGES,
            num_secondary_workers=scatter_count if preempt else 0,
            num_workers=2 if preempt else scatter_count,
            cluster_name=name,
            depends_on=depends_on,
            init=['gs://cpg-reference/hail_dataproc/install_phantomjs.sh']
            if phantomjs
            else [],
        )
    return clusters_by_name[name]



#!/usr/bin/env python

"""
Runs PCA, optionally on one population specifically
"""

import logging
from typing import Optional

import click
import hail as hl

from joint_calling import utils
from joint_calling import _version
from joint_calling import sample_qc as sqc


logger = logging.getLogger(__file__)
logging.basicConfig(format='%(levelname)s (%(name)s %(lineno)s): %(message)s')
logger.setLevel(logging.INFO)


@click.command()
@click.version_option(_version.__version__)
@click.option(
    '--hgdp-union-mt',
    'hgdp_union_mt_path',
    required=True,
    callback=utils.get_validation_callback(ext='mt', must_exist=True),
    help='path to Matrix Table generated with sample_qc_subset_mt_for_pca.py',
)
@click.option(
    '--n-pcs', 'n_pcs', type=int, help='number of PCs to compute for ancestry PCA.'
)
@click.option(
    '--pop',
    'pop',
    help='population label to subset the training dataset to',
)
@click.option(
    '--related-samples-to-drop-ht',
    'related_samples_to_drop_ht_path',
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
    help='Samples to remove from the analysis',
)
@click.option(
    '--out-eigenvalues',
    'out_eigenvalues_path',
    callback=utils.get_validation_callback(ext='txt', must_exist=False),
)
@click.option(
    '--out-scores-ht',
    'out_scores_ht_path',
    callback=utils.get_validation_callback(ext='ht', must_exist=False),
)
@click.option(
    '--out-loadings-ht',
    'out_loadings_ht_path',
    callback=utils.get_validation_callback(ext='ht', must_exist=False),
)
@click.option(
    '--tmp-bucket',
    'tmp_bucket',
    required=True,
    help='path to write temporary intermediate output and checkpoints',
)
@click.option(
    '--overwrite/--reuse',
    'overwrite',
    is_flag=True,
    help='if an intermediate or a final file exists, skip running the code '
    'that generates it.',
)
@click.option(
    '--hail-billing',
    'hail_billing',
    required=True,
    help='Hail billing account ID.',
)
def main(  # pylint: disable=too-many-arguments,too-many-locals,missing-function-docstring
    hgdp_union_mt_path: str,
    n_pcs: int,
    related_samples_to_drop_ht_path: Optional[str],
    pop: Optional[str],
    out_eigenvalues_path: str,
    out_scores_ht_path: str,
    out_loadings_ht_path: str,
    tmp_bucket: str,
    overwrite: bool,
    hail_billing: str,  # pylint: disable=unused-argument
):
    utils.init_hail(__file__)

    if all(
        utils.can_reuse(fp, overwrite)
        for fp in [out_eigenvalues_path, out_scores_ht_path, out_loadings_ht_path]
    ):
        return

    mt = hl.read_matrix_table(hgdp_union_mt_path)

    if pop:
        # Get samples from the specified population only
        mt = mt.filter_cols(
            ~hl.is_defined(mt.hgdp_1kg_metadata)
            | (mt.hgdp_1kg_metadata.population_inference.pop == pop.lower())
        )

    related_samples_to_drop_ht = None
    if related_samples_to_drop_ht_path:
        related_samples_to_drop_ht = hl.read_table(related_samples_to_drop_ht_path)

    sqc.run_pca_ancestry_analysis(
        mt=mt,
        sample_to_drop_ht=related_samples_to_drop_ht,
        tmp_bucket=tmp_bucket,
        n_pcs=n_pcs,
        out_eigenvalues_path=out_eigenvalues_path,
        out_scores_ht_path=out_scores_ht_path,
        out_loadings_ht_path=out_loadings_ht_path,
        overwrite=overwrite,
    )


if __name__ == '__main__':
    main()  # pylint: disable=E1120



#!/usr/bin/env python

"""
Computes a HT with the typical GATK AS and site-level info fields
as well as ACs and lowqual fields. Note that this table doesn't 
split multi-allelic sites.

Generates info.ht, info-split.ht, needed for sample_qc and for 
random forest jobs.
"""

import logging
import click
import hail as hl

from gnomad.utils.annotations import (
    get_adj_expr,
    get_lowqual_expr,
)
from gnomad.utils.sparse_mt import (
    get_as_info_expr,
    get_site_info_expr,
    INFO_INT32_SUM_AGG_FIELDS,
    INFO_SUM_AGG_FIELDS,
    split_info_annotation,
    split_lowqual_annotation,
)

from joint_calling.utils import file_exists
from joint_calling import utils, _version

logger = logging.getLogger('qc-annotations')
logger.setLevel(logging.INFO)


@click.command()
@click.version_option(_version.__version__)
@click.option(
    '--out-info-ht',
    'out_info_ht_path',
    required=True,
)
@click.option(
    '--out-split-info-ht',
    'out_split_info_ht_path',
    required=True,
)
@click.option(
    '--mt',
    'mt_path',
    required=True,
    callback=utils.get_validation_callback(ext='mt', must_exist=True),
    help='path to the input MatrixTable',
)
@click.option(
    '--local-tmp-dir',
    'local_tmp_dir',
    help='local directory for temporary files and Hail logs (must be local).',
)
@click.option(
    '--overwrite/--reuse',
    'overwrite',
    is_flag=True,
    help='if an intermediate or a final file exists, skip running the code '
    'that generates it.',
)
def main(  # pylint: disable=too-many-arguments,too-many-locals,too-many-statements,missing-function-docstring
    out_info_ht_path: str,
    out_split_info_ht_path: str,
    mt_path: str,
    local_tmp_dir: str,
    overwrite: bool,
):
    local_tmp_dir = utils.init_hail('generate_info_ht', local_tmp_dir)

    all_samples_mt = utils.get_mt(mt_path)

    compute_info(
        mt=all_samples_mt,
        out_ht_path=out_info_ht_path,
        out_split_ht_path=out_split_info_ht_path,
        overwrite=overwrite,
    )


def compute_info(
    mt: hl.MatrixTable,
    out_ht_path: str,
    out_split_ht_path: str,
    overwrite: bool = False,
) -> hl.Table:
    """
    :param mt: full matrix table
    :param out_ht_path: where to write the info Table
    :param out_split_ht_path: if provided, in the info Table multiallelics will be split
    and the Table will be written to this file
    :param overwrite: overwrite checkpoints if they exist
    :return: Table with info fields
    """
    if not overwrite and file_exists(out_ht_path):
        info_ht = hl.read_table(out_ht_path)
    else:
        mt = mt.filter_rows((hl.len(mt.alleles) > 1))
        mt = mt.transmute_entries(**mt.gvcf_info)
        mt = mt.annotate_rows(alt_alleles_range_array=hl.range(1, hl.len(mt.alleles)))

        # Compute AS and site level info expr
        # Note that production defaults have changed:
        # For new releases, the `RAW_MQandDP` field replaces the `RAW_MQ` and `MQ_DP` fields
        info_expr = get_site_info_expr(
            mt,
            sum_agg_fields=INFO_SUM_AGG_FIELDS,
            int32_sum_agg_fields=INFO_INT32_SUM_AGG_FIELDS,
            array_sum_agg_fields=['SB', 'RAW_MQandDP'],
        )
        info_expr = info_expr.annotate(
            **get_as_info_expr(
                mt,
                sum_agg_fields=INFO_SUM_AGG_FIELDS,
                int32_sum_agg_fields=INFO_INT32_SUM_AGG_FIELDS,
                array_sum_agg_fields=['SB', 'RAW_MQandDP'],
            )
        )

        # Add AC and AC_raw:
        # First compute ACs for each non-ref allele, grouped by adj
        grp_ac_expr = hl.agg.array_agg(
            lambda ai: hl.agg.filter(
                mt.LA.contains(ai),
                hl.agg.group_by(
                    get_adj_expr(mt.LGT, mt.GQ, mt.DP, mt.LAD),
                    hl.agg.sum(
                        mt.LGT.one_hot_alleles(mt.LA.map(hl.str))[mt.LA.index(ai)]
                    ),
                ),
            ),
            mt.alt_alleles_range_array,
        )

        # Then, for each non-ref allele, compute
        # AC as the adj group
        # AC_raw as the sum of adj and non-adj groups
        info_expr = info_expr.annotate(
            AC_raw=grp_ac_expr.map(
                lambda i: hl.int32(i.get(True, 0) + i.get(False, 0))
            ),
            AC=grp_ac_expr.map(lambda i: hl.int32(i.get(True, 0))),
        )

        # Annotating raw MT with pab max
        info_expr = info_expr.annotate(
            AS_pab_max=hl.agg.array_agg(
                lambda ai: hl.agg.filter(
                    mt.LA.contains(ai) & mt.LGT.is_het(),
                    hl.agg.max(
                        hl.binom_test(
                            mt.LAD[mt.LA.index(ai)], hl.sum(mt.LAD), 0.5, 'two-sided'
                        )
                    ),
                ),
                mt.alt_alleles_range_array,
            )
        )

        info_ht = mt.select_rows(info=info_expr).rows()

        # Add lowqual flag
        info_ht = info_ht.annotate(
            lowqual=get_lowqual_expr(
                info_ht.alleles,
                info_ht.info.QUALapprox,
                # The indel het prior used for gnomad v3 was 1/10k bases (phred=40).
                # This value is usually 1/8k bases (phred=39).
                indel_phred_het_prior=40,
            ),
            AS_lowqual=get_lowqual_expr(
                info_ht.alleles, info_ht.info.AS_QUALapprox, indel_phred_het_prior=40
            ),
        )
        info_ht = info_ht.naive_coalesce(7500)
        info_ht.write(out_ht_path, overwrite=True)

    if out_split_ht_path and (overwrite or not file_exists(out_split_ht_path)):
        split_info_ht = split_multiallelic_in_info_table(info_ht)
        split_info_ht.write(out_split_ht_path, overwrite=True)

    return info_ht


def split_multiallelic_in_info_table(info_ht: hl.Table) -> hl.Table:
    """
    Generates an info table that splits multi-allelic sites from the multi-allelic
    info table.
    :return: Info table with split multi-allelics
    :rtype: Table
    """

    # Create split version
    info_ht = hl.split_multi(info_ht)

    info_ht = info_ht.annotate(
        info=info_ht.info.annotate(
            **split_info_annotation(info_ht.info, info_ht.a_index),
        ),
        AS_lowqual=split_lowqual_annotation(info_ht.AS_lowqual, info_ht.a_index),
    )
    return info_ht


if __name__ == '__main__':
    main()  # pylint: disable=E1120



"""
Just defines `package_path` which returns the local install path of the package
"""

from os.path import dirname, abspath


def get_package_path():
    """
    :return: local install path of the package
    """
    return dirname(abspath(__file__))



#!/usr/bin/env python

"""
Flag related samples to further exclude from the PCA analysis
"""

import logging
import click
import hail as hl

from joint_calling import sample_qc as sqc, utils
from joint_calling import _version


logger = logging.getLogger(__file__)
logging.basicConfig(format='%(levelname)s (%(name)s %(lineno)s): %(message)s')
logger.setLevel(logging.INFO)


@click.command()
@click.version_option(_version.__version__)
@click.option(
    '--hard-filtered-samples-ht',
    'hard_filtered_samples_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
)
@click.option(
    '--sex-ht',
    'sex_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
    help='Just for the `chr20_mean_dp` column',
)
@click.option(
    '--relatedness-ht',
    'relatedness_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
)
@click.option(
    '--out-ht',
    'out_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=False),
)
@click.option(
    '--tmp-bucket',
    'tmp_bucket',
    required=True,
    help='path to write temporary intermediate output and checkpoints',
)
@click.option(
    '--max-kin',
    'max_kin',
    type=float,
)
@click.option(
    '--overwrite/--reuse',
    'overwrite',
    is_flag=True,
    help='if an intermediate or a final file exists, skip running the code '
    'that generates it.',
)
@click.option(
    '--hail-billing',
    'hail_billing',
    required=True,
    help='Hail billing account ID.',
)
def main(  # pylint: disable=too-many-arguments,too-many-locals,missing-function-docstring
    hard_filtered_samples_ht_path: str,
    sex_ht_path: str,
    relatedness_ht_path: str,
    out_ht_path: str,
    max_kin: int,
    tmp_bucket: str,
    overwrite: bool,
    hail_billing: str,  # pylint: disable=unused-argument
):
    if utils.can_reuse(out_ht_path, overwrite):
        return

    utils.init_hail(__file__)
    sqc.flag_related_samples(
        hard_filtered_samples_ht=hl.read_table(hard_filtered_samples_ht_path),
        sex_ht=hl.read_table(sex_ht_path),
        relatedness_ht=hl.read_table(relatedness_ht_path),
        regressed_metrics_ht=None,
        tmp_bucket=tmp_bucket,
        kin_threshold=max_kin,
        overwrite=overwrite,
        out_ht_path=out_ht_path,
    )


if __name__ == '__main__':
    main()  # pylint: disable=E1120



"""Functions to infer ancestry and perform ancestry-stratfied sample QC"""

import logging
import pickle
from os.path import join
from typing import Optional, List, Dict
import hail as hl
import pandas as pd

from gnomad.sample_qc.ancestry import run_pca_with_relateds, assign_population_pcs
from gnomad.sample_qc.relatedness import compute_related_samples_to_drop
from gnomad.sample_qc.filtering import (
    compute_qc_metrics_residuals,
    compute_stratified_metrics_filter,
)
from gnomad.sample_qc.filtering import compute_stratified_sample_qc
from gnomad.sample_qc.pipeline import annotate_sex
from gnomad.utils.file_utils import file_exists
from gnomad.utils.annotations import bi_allelic_expr
from gnomad.utils.filtering import filter_to_autosomes
from gnomad.utils.sparse_mt import filter_ref_blocks

from joint_calling import utils


logger = logging.getLogger(__file__)
logging.basicConfig(format='%(levelname)s (%(name)s %(lineno)s): %(message)s')
logger.setLevel(logging.INFO)


def compute_hail_sample_qc(
    mt: hl.MatrixTable,
    tmp_bucket: str,
    out_ht_path: Optional[str] = None,
    overwrite: bool = False,
) -> hl.Table:
    """
    Runs Hail hl.sample_qc() to generate a Hail Table with
    per-sample QC metrics.
    :param mt: input MatrixTable, multiallelics split, genotype in GT fields
    :param tmp_bucket: bucket path to write checkpoints
    :param out_ht_path: location to write the result to
    :param overwrite: overwrite checkpoints if they exist
    :return: a Hail Table with the following row fields:
        'sample_qc': {
            <output of hl.sample_qc() (n_filtered, n_hom_ref, etc)>
        }
        'bi_allelic_sample_qc': {
            <output of hl.sample_qc() for bi-allelic variants only>
        }
        'multi_allelic_sample_qc': {
            <output of hl.sample_qc() for multi-allelic variants only>
        }
    """
    logger.info('Sample QC')
    out_ht_path = out_ht_path or join(tmp_bucket, 'hail_sample_qc.ht')
    if utils.can_reuse(out_ht_path, overwrite):
        return hl.read_table(out_ht_path)

    mt = filter_to_autosomes(mt)

    mt = mt.select_entries('GT')

    mt = filter_ref_blocks(mt)

    # Remove centromeres and telomeres incase they were included and any reference blocks
    tel_cent_ht = hl.read_table(utils.TEL_AND_CENT_HT_PATH)
    mt = mt.filter_rows(hl.is_missing(tel_cent_ht[mt.locus]))

    sample_qc_ht = compute_stratified_sample_qc(
        mt,
        strata={
            'bi_allelic': bi_allelic_expr(mt),
            'multi_allelic': ~bi_allelic_expr(mt),
        },
        tmp_ht_prefix=None,
    )

    # Remove annotations that cannot be computed from the sparse format
    sample_qc_ht = sample_qc_ht.annotate(
        **{
            x: sample_qc_ht[x].drop('n_called', 'n_not_called', 'call_rate')
            for x in sample_qc_ht.row_value
        }
    )
    sample_qc_ht = sample_qc_ht.repartition(100)
    return sample_qc_ht.checkpoint(out_ht_path, overwrite=True)


def snps_not_in_gnomad(
    mt: hl.MatrixTable,
    tmp_bucket: str,
    out_ht_path: Optional[str] = None,
    overwrite: bool = False,
    gnomad_path: str = utils.GNOMAD_HT_PATH,
) -> hl.Table:
    """
    Count the number of variants per sample that do not occur in gnomAD
    :param mt: MatrixTable, with multiallelics split, GT field for genotype
    :param tmp_bucket: bucket path to write checkpoints
    :param out_ht_path: location to write the result to
    :param overwrite: overwrite checkpoints if they exist
    :param gnomad_path: path to GnomAD Hail Table
    :return: per-sample Table, with the only int field "nongnomad_snps"
    """
    # Filter to SNPs
    logger.info('Countings SNPs not in gnomAD')
    out_ht_path = out_ht_path or join(tmp_bucket, 'notingnomad.ht')
    if utils.can_reuse(out_ht_path, overwrite):
        return hl.read_table(out_ht_path)

    mt = mt.filter_rows(hl.len(mt.alleles) > 1)
    mt = mt.filter_rows(hl.is_snp(mt.alleles[0], mt.alleles[1]))

    # Get entries (table annotated with locus, allele, sample)
    ht = mt.entries()

    # Filter to those variants that are not in gnomad
    gnomad = hl.read_table(gnomad_path)
    ht = ht.key_by('locus', 'alleles').anti_join(gnomad)
    # Count non-gnomad variants for each sample
    stats_ht = ht.group_by(ht.s).aggregate(nongnomad_snps=hl.agg.count())
    return stats_ht.checkpoint(out_ht_path, overwrite=True)


def infer_sex(
    mt: hl.MatrixTable,
    tmp_bucket: str,
    out_ht_path: Optional[str] = None,
    overwrite: bool = False,
    target_regions: Optional[hl.Table] = None,
) -> hl.Table:
    """
    Runs gnomad.sample_qc.annotate_sex() to infer sex
    :param mt: input MatrixTable
    :param tmp_bucket: bucket to write checkpoints
    :param out_ht_path: location to write the result to
    :param overwrite: overwrite checkpoints if they exist
    :param target_regions: if exomes, it should correspond to the regions in the panel
    :return: a Table with the following row fields:
        'is_female': bool
        'f_stat': float64
        'n_called': int64
        'expected_homs': float64
        'observed_homs': int64
        'chr20_mean_dp': float32
        'chrX_mean_dp': float32
        'chrY_mean_dp': float32
        'chrX_ploidy': float32
        'chrY_ploidy': float32
        'X_karyotype': str
        'Y_karyotype': str
        'sex_karyotype': str
    """
    logger.info('Inferring sex')
    out_ht_path = out_ht_path or join(tmp_bucket, 'sex.ht')
    if utils.can_reuse(out_ht_path, overwrite):
        return hl.read_table(out_ht_path)

    ht = annotate_sex(
        mt,
        excluded_intervals=hl.read_table(utils.TEL_AND_CENT_HT_PATH),
        included_intervals=target_regions,
        gt_expr='LGT',
    )

    return ht.checkpoint(out_ht_path, overwrite=True)


def run_pca_ancestry_analysis(
    mt: hl.MatrixTable,
    sample_to_drop_ht: Optional[hl.Table],
    tmp_bucket: str,
    n_pcs: int,
    out_eigenvalues_path: Optional[str] = None,
    out_scores_ht_path: Optional[str] = None,
    out_loadings_ht_path: Optional[str] = None,
    overwrite: bool = False,
) -> hl.Table:
    """
    :param mt: variants usable for PCA analysis, combined with samples
        with known populations (HGDP, 1KG, etc)
    :param tmp_bucket: bucket path to write checkpoints
    :param sample_to_drop_ht: table with samples to drop based on
        previous relatedness analysis. With a `rank` row field
    :param n_pcs: maximum number of principal components
    :param overwrite: overwrite checkpoints if they exist
    :param out_eigenvalues_path: path to a txt file to write PCA eigenvalues
    :param out_scores_ht_path: path to write PCA scores
    :param out_loadings_ht_path: path to write PCA loadings
    :return: a Table with a row field:
        'scores': array<float64>
    """
    logger.info('Running PCA ancestry analysis')
    out_scores_ht_path = out_scores_ht_path or join(tmp_bucket, 'pop_pca_scores.ht')
    if all(
        not fp or utils.can_reuse(fp, overwrite)
        for fp in [
            out_eigenvalues_path,
            out_scores_ht_path,
            out_loadings_ht_path,
        ]
    ):
        return hl.read_table(out_scores_ht_path)

    # Adjusting the number of principal components not to exceed the
    # number of samples
    samples_to_drop_num = 0 if sample_to_drop_ht is None else sample_to_drop_ht.count()
    n_pcs = min(n_pcs, mt.cols().count() - samples_to_drop_num)
    eigenvalues, scores_ht, loadings_ht = run_pca_with_relateds(
        mt, sample_to_drop_ht, n_pcs=n_pcs
    )

    if out_eigenvalues_path:
        hl.Table.from_pandas(pd.DataFrame(eigenvalues)).export(out_eigenvalues_path)
    if out_loadings_ht_path:
        loadings_ht.write(out_loadings_ht_path, overwrite=True)
    scores_ht.write(out_scores_ht_path, overwrite=True)
    scores_ht = hl.read_table(out_scores_ht_path)
    return scores_ht


def infer_pop_labels(
    pop_pca_scores_ht: hl.Table,
    provided_pop_ht: hl.Table,
    tmp_bucket: str,
    min_prob: float,
    max_mislabeled_training_samples: int = 50,
    n_pcs: int = 16,
    out_ht_path: Optional[str] = None,
    overwrite: bool = False,
) -> hl.Table:
    """
    Take population PCA results and training data, and run random forest
    to assign global population labels.

    :param pop_pca_scores_ht: output table of `_run_pca_ancestry_analysis()`
        with a row field 'scores': array<float64>
    :param provided_pop_ht: table with a `continental_pop` field. Samples for which
        the latter is defined will be used to train the random forest
    :param tmp_bucket: bucket to write checkpoints and intermediate files
    :param min_prob: min probability of belonging to a given population
        for the population to be set (otherwise set to `None`)
    :param max_mislabeled_training_samples: keep rerunning until the number
        of mislabeled samples is below this number
    :param n_pcs: Number of PCs to use in the RF
    :param out_ht_path: Path to write the resulting HT table
    :param overwrite: overwrite checkpoints if they exist
    :return: a Table with the following row fields, including `prob_<POP>`
        probabily fields for each population label:
        'training_pop': str
        'pca_scores': array<float64>
        'pop': str
        'prob_CEU': float64
        'prob_YRI': float64
        ... (prob_*: float64 for each population label)
    """
    logger.info('Assigning global population labels')
    out_ht_path = out_ht_path or join(tmp_bucket, 'inferred_pop.ht')
    if utils.can_reuse(out_ht_path, overwrite):
        return hl.read_table(out_ht_path)

    samples_with_pop_ht = provided_pop_ht.filter(
        hl.is_defined(provided_pop_ht.continental_pop)
        & (provided_pop_ht.continental_pop != '')
    )
    pop_pca_scores_ht = pop_pca_scores_ht.annotate(
        training_pop=samples_with_pop_ht[pop_pca_scores_ht.key].continental_pop
    )

    def _run_assign_population_pcs(pop_pca_scores_ht, min_prob):
        examples_num = pop_pca_scores_ht.aggregate(
            hl.agg.count_where(hl.is_defined(pop_pca_scores_ht.training_pop))
        )
        logger.info(f'Running RF using {examples_num} training examples')
        pop_ht, pops_rf_model = assign_population_pcs(
            pop_pca_scores_ht,
            pc_cols=pop_pca_scores_ht.scores[:n_pcs],
            known_col='training_pop',
            min_prob=min_prob,
        )
        n_mislabeled_samples = pop_ht.aggregate(
            hl.agg.count_where(pop_ht.training_pop != pop_ht.pop)
        )
        return pop_ht, pops_rf_model, n_mislabeled_samples

    pop_ht, pops_rf_model, n_mislabeled_samples = _run_assign_population_pcs(
        pop_pca_scores_ht, min_prob
    )
    while n_mislabeled_samples > max_mislabeled_training_samples:
        logger.info(
            f'Found {n_mislabeled_samples} samples '
            f'labeled differently from their known pop. '
            f'Re-running without them.'
        )

        pop_ht = pop_ht[pop_pca_scores_ht.key]
        pop_pca_scores_ht = pop_pca_scores_ht.annotate(
            training_pop=hl.or_missing(
                (pop_ht.training_pop == pop_ht.pop), pop_pca_scores_ht.training_pop
            )
        ).persist()

        pop_ht, pops_rf_model, n_mislabeled_samples = _run_assign_population_pcs(
            pop_pca_scores_ht, min_prob
        )

    # Writing a tab delimited file indicating inferred sample populations
    pop_tsv_file = join(tmp_bucket, 'RF_pop_assignments.txt.gz')
    if overwrite or not file_exists(pop_tsv_file):
        pc_cnt = min(hl.min(10, hl.len(pop_ht.pca_scores)).collect())
        pop_ht.transmute(
            **{f'PC{i + 1}': pop_ht.pca_scores[i] for i in range(pc_cnt)}
        ).export(pop_tsv_file)

    # Writing the RF model used for inferring sample populations
    pop_rf_file = join(tmp_bucket, 'pop.RF_fit.pickle')
    if overwrite or not file_exists(pop_rf_file):
        with hl.hadoop_open(pop_rf_file, 'wb') as out:
            pickle.dump(pops_rf_model, out)

    return pop_ht.checkpoint(out_ht_path, overwrite=True)


def compute_stratified_qc(
    sample_qc_ht: hl.Table,
    pop_ht: hl.Table,
    work_bucket: str,
    filtering_qc_metrics: List[str],
    overwrite: bool = False,
) -> hl.Table:
    """
    Computes median, MAD, and upper and lower thresholds for each metric
    in `filtering_qc_metrics`, groupped by `pop` field in `pop_ht`

    :param sample_qc_ht: table with a row field
        `sample_qc` = struct { n_snp: int64, n_singleton: int64, ... }
    :param pop_ht: table with a `pop` row field
    :param work_bucket: bucket to write checkpoints
    :param filtering_qc_metrics: metrics to annotate with
    :param overwrite: overwrite checkpoints if they exist
    :return: table with the following structure:
        Global fields:
            'qc_metrics_stats': dict<tuple (
                str
                ), struct {
                    n_snp: struct {
                        median: int64,
                        mad: float64,
                        lower: float64,
                        upper: float64
                    },
                    n_singleton: struct {
                        ...
                    },
                    ...
            }
        Row fields:
            'fail_n_snp': bool
            'fail_n_singleton': bool
            ...
            'qc_metrics_filters': set<str>
    """
    logger.info(
        f'Computing stratified QC metrics filters using '
        f'metrics: {", ".join(filtering_qc_metrics)}'
    )

    sample_qc_ht = sample_qc_ht.annotate(qc_pop=pop_ht[sample_qc_ht.key].pop)
    stratified_metrics_ht = compute_stratified_metrics_filter(
        sample_qc_ht,
        qc_metrics={
            metric: sample_qc_ht.sample_qc[metric] for metric in filtering_qc_metrics
        },
        strata={'qc_pop': sample_qc_ht.qc_pop},
        metric_threshold={'n_singleton': (4.0, 8.0)},
    )
    return stratified_metrics_ht.checkpoint(
        join(work_bucket, 'stratified_metrics.ht'),
        overwrite=overwrite,
        _read_if_exists=not overwrite,
    )


def flag_related_samples(
    hard_filtered_samples_ht: hl.Table,
    sex_ht: hl.Table,
    relatedness_ht: hl.Table,
    regressed_metrics_ht: Optional[hl.Table],
    tmp_bucket: str,
    kin_threshold: float,
    out_ht_path: Optional[str] = None,
    overwrite: bool = False,
) -> hl.Table:
    """
    Flag samples to drop based on relatedness, so the final set
    has only unrelated samples, best quality one per family

    :param hard_filtered_samples_ht: table with failed samples
        and a `hard_filters` row field
    :param sex_ht: table with a `chr20_mean_dp` row field
    :param relatedness_ht: table keyed by exactly two fields (i and j)
        of the same type, identifying the pair of samples for each row
    :param regressed_metrics_ht: optional table with a `qc_metrics_filters`
        field calculated with _apply_regressed_filters() from PCA scores
    :param tmp_bucket: bucket to write checkpoints
    :param kin_threshold: kinship threshold to call two samples as related
    :param overwrite: overwrite checkpoints if they exist
    :param out_ht_path: path to write the resulting Table to
    :return: a table of the samples to drop along with their rank
        row field: 'rank': int64
    """
    label = 'final' if regressed_metrics_ht is not None else 'intermediate'
    logger.info(f'Flagging related samples to drop, {label}')
    out_ht_path = out_ht_path or join(tmp_bucket, f'{label}_related_samples_to_drop.ht')
    if utils.can_reuse(out_ht_path, overwrite):
        return hl.read_table(out_ht_path)

    rankings_ht_path = join(tmp_bucket, f'{label}_samples_rankings.ht')
    if utils.can_reuse(rankings_ht_path, overwrite):
        rank_ht = hl.read_table(rankings_ht_path)
    else:
        rank_ht = _compute_sample_rankings(
            hard_filtered_samples_ht,
            sex_ht,
            use_qc_metrics_filters=regressed_metrics_ht is not None,
            regressed_metrics_ht=regressed_metrics_ht,
        ).checkpoint(rankings_ht_path, overwrite=True)

    try:
        filtered_samples = hl.literal(
            rank_ht.aggregate(
                hl.agg.filter(rank_ht.filtered, hl.agg.collect(rank_ht.s))
            )
        )
    except hl.ExpressionException:
        # Hail doesn't handle it with `aggregate` when none of
        # the samples is 'filtered'
        filtered_samples = hl.empty_array('tstr')

    samples_to_drop_ht = compute_related_samples_to_drop(
        relatedness_ht,
        rank_ht,
        kin_threshold=kin_threshold,
        filtered_samples=filtered_samples,
    )
    return samples_to_drop_ht.checkpoint(out_ht_path, overwrite=True)


def _compute_sample_rankings(
    hard_filtered_samples_ht: hl.Table,
    sex_ht: hl.Table,
    use_qc_metrics_filters: bool = False,
    regressed_metrics_ht: Optional[hl.Table] = None,
) -> hl.Table:
    """
    Orders samples by hard filters and coverage and adds rank,
    which is the lower the better.

    :param hard_filtered_samples_ht: table with failed samples
        and a `hard_filters` row field
    :param sex_ht: table with a `chr20_mean_dp` row field
    :param use_qc_metrics_filters: apply population-stratified QC filters
    :param regressed_metrics_ht: table with a `qc_metrics_filters` field.
        Used only if `use_qc_metrics_filters` is True.
    :return: table ordered by rank, with the following row fields:
        `rank`, `filtered`
    """
    ht = sex_ht.drop(*list(sex_ht.globals.dtype.keys()))
    ht = ht.select(
        'chr20_mean_dp',
        filtered=hl.or_else(
            hl.len(hard_filtered_samples_ht[ht.key].hard_filters) > 0, False
        ),
    )
    if use_qc_metrics_filters and regressed_metrics_ht is not None:
        ht = ht.annotate(
            filtered=hl.cond(
                ht.filtered,
                True,
                hl.or_else(
                    hl.len(regressed_metrics_ht[ht.key].qc_metrics_filters) > 0, False
                ),
            )
        )

    ht = ht.order_by(ht.filtered, hl.desc(ht.chr20_mean_dp)).add_index(name='rank')
    return ht.key_by('s').select('filtered', 'rank')


def apply_regressed_filters(
    sample_qc_ht: hl.Table,
    pop_pca_scores_ht: hl.Table,
    tmp_bucket: str,
    out_ht_path: Optional[str] = None,
    overwrite: bool = False,
) -> hl.Table:
    """
    Re-compute QC metrics (with hl.sample_qc() - like n_snp, r_het_hom)
    per population, and adding "fail_*" row fields when a metric is below
    the the lower MAD threshold or higher the upper MAD threshold
    (see `compute_stratified_metrics_filter` for defaults)

    :param sample_qc_ht: table with a row field
       `bi_allelic_sample_qc` =
          struct { n_snp: int64, n_singleton: int64, ... }
    :param pop_pca_scores_ht: table with a `scores` row field
    :param tmp_bucket: bucket to write checkpoints
    :param out_ht_path: path to write the resulting table to
    :param overwrite: overwrite checkpoints if they exist
    :return: a table with the folliwing structure:
        Global fields:
            'lms': struct {
                n_snp: struct {
                    beta: array<float64>,
                    standard_error: array<float64>,
                    t_stat: array<float64>,
                    p_value: array<float64>,
                    multiple_standard_error: float64,
                    multiple_r_squared: float64,
                    adjusted_r_squared: float64,
                    f_stat: float64,
                    multiple_p_value: float64,
                    n: int32
                },
                n_singleton: struct {
                    ...
                },
                ...
            }
            'qc_metrics_stats': struct {
                n_snp_residual: struct {
                    median: float64,
                    mad: float64,
                    lower: float64,
                    upper: float64
                },
                n_singleton_residual: struct {
                    ...
                },
                ...
            }
        Row fields:
            's': str
            'n_snp_residual': float64
            'n_singleton_residual': float64
            ...
            'fail_n_snp_residual': bool
            'fail_n_singleton_residual': bool
            ...
            'qc_metrics_filters': set<str>
    """
    logger.info('Compute QC metrics adjusted for popopulation')
    out_ht_path = out_ht_path or join(tmp_bucket, 'regressed_metrics.ht')
    if utils.can_reuse(out_ht_path, overwrite):
        return hl.read_table(out_ht_path)

    sample_qc_ht = sample_qc_ht.select(
        **sample_qc_ht.bi_allelic_sample_qc,
        **pop_pca_scores_ht[sample_qc_ht.key],
        releasable=hl.bool(True),
    )

    filtering_qc_metrics = [
        'n_snp',
        'n_singleton',
        'r_ti_tv',
        'r_insertion_deletion',
        'n_insertion',
        'n_deletion',
        'r_het_hom_var',
        'n_het',
        'n_hom_var',
        'n_transition',
        'n_transversion',
    ]
    residuals_ht = compute_qc_metrics_residuals(
        ht=sample_qc_ht,
        pc_scores=sample_qc_ht.scores,
        qc_metrics={metric: sample_qc_ht[metric] for metric in filtering_qc_metrics},
        regression_sample_inclusion_expr=sample_qc_ht.releasable,
    )

    stratified_metrics_ht = compute_stratified_metrics_filter(
        ht=residuals_ht,
        qc_metrics=dict(residuals_ht.row_value),
        metric_threshold={'n_singleton_residual': (4.0, 8.0)},
    )

    residuals_ht = residuals_ht.annotate(**stratified_metrics_ht[residuals_ht.key])
    residuals_ht = residuals_ht.annotate_globals(
        **stratified_metrics_ht.index_globals()
    )

    return residuals_ht.checkpoint(out_ht_path, overwrite=True)


def compute_hard_filters(
    mt: hl.MatrixTable,
    picard_metrics_ht: hl.Table,
    sex_ht: hl.Table,
    hail_sample_qc_ht: hl.Table,
    cutoffs_d: Dict,
    out_ht_path: str,
    overwrite: bool = False,
) -> hl.Table:
    """
    Uses the sex imputation results, results of the sample_qc() run on
    bi-allelic variants, and stats by Picard tools specificed in `sample_df`,
    to apply filters to samples in `mt`, and create a table with
    only samples that fail at least one sample.

    :param mt: input matrix table
    :param picard_metrics_ht: table QC metrics from Picard tools. Expected
        fields: r_contamination, r_chimera, r_duplication, median_insert_size
    :param sex_ht: required fields: "sex_karyotype", "chr20_mean_dp"
    :param hail_sample_qc_ht: required fields:
        "bi_allelic_sample_qc { n_snp, n_singleton, r_het_hom_var }"
    :param out_ht_path: location to write the hard filtered samples Table
    :param cutoffs_d: a dictionary with hard-filtering thresholds
    :param overwrite: overwrite checkpoints if they exist
    :return: a table with samples failed the filters, and the following structure:
        's': str
        'hard_filters': set<str>  # a non-empty subset of { ambiguous_sex,
            sex_aneuploidy,  low_coverage, bad_biallelic_metrics, contamination,
            chimera, coverage, insert_size }
    """
    logger.info('Generating hard filters')
    if utils.can_reuse(out_ht_path, overwrite):
        return hl.read_table(out_ht_path)

    ht = mt.cols()
    ht = ht.annotate(hard_filters=hl.empty_set(hl.tstr))

    # Helper function to add filters into the `hard_filters` set
    def add_filter(ht, expr, name):
        return ht.annotate(
            hard_filters=hl.if_else(
                expr & hl.is_defined(expr), ht.hard_filters.add(name), ht.hard_filters
            )
        )

    # Remove samples with ambiguous sex assignments
    ht = add_filter(ht, sex_ht[ht.key].sex_karyotype == 'ambiguous', 'ambiguous_sex')
    ht = add_filter(
        ht,
        ~hl.set({'ambiguous', 'XX', 'XY'}).contains(sex_ht[ht.key].sex_karyotype),
        'sex_aneuploidy',
    )

    # Remove low-coverage samples
    # chrom 20 coverage is computed to infer sex and used here
    ht = add_filter(
        ht, sex_ht[ht.key].chr20_mean_dp < cutoffs_d['min_coverage'], 'low_coverage'
    )

    # Remove extreme raw bi-allelic sample QC outliers
    ht = add_filter(
        ht,
        (
            (
                hail_sample_qc_ht[ht.key].bi_allelic_sample_qc.n_snp
                > cutoffs_d['max_n_snps']
            )
            | (
                hail_sample_qc_ht[ht.key].bi_allelic_sample_qc.n_snp
                < cutoffs_d['min_n_snps']
            )
            | (
                hail_sample_qc_ht[ht.key].bi_allelic_sample_qc.n_singleton
                > cutoffs_d['max_n_singletons']
            )
            | (
                hail_sample_qc_ht[ht.key].bi_allelic_sample_qc.r_het_hom_var
                > cutoffs_d['max_r_het_hom']
            )
        ),
        'bad_biallelic_metrics',
    )

    ht = add_filter(
        ht,
        hl.is_missing(picard_metrics_ht[ht.key].r_contamination)
        | (
            picard_metrics_ht[ht.key].r_contamination > cutoffs_d['max_r_contamination']
        ),
        'contamination',
    )
    ht = add_filter(
        ht,
        hl.is_missing(picard_metrics_ht[ht.key].r_chimera)
        | (picard_metrics_ht[ht.key].r_chimera > cutoffs_d['max_r_chimera']),
        'chimera',
    )
    ht = add_filter(
        ht,
        hl.is_missing(picard_metrics_ht[ht.key].r_duplication)
        | (picard_metrics_ht[ht.key].r_duplication > cutoffs_d['max_r_duplication']),
        'dup_rate',
    )
    ht = add_filter(
        ht,
        hl.is_missing(picard_metrics_ht[ht.key].median_insert_size)
        | (
            picard_metrics_ht[ht.key].median_insert_size
            < cutoffs_d['min_median_insert_size']
        ),
        'insert_size',
    )
    ht = ht.annotate_globals(hard_filter_cutoffs=hl.struct(**cutoffs_d))
    ht = ht.filter(hl.len(ht.hard_filters) > 0)
    return ht.checkpoint(out_ht_path, overwrite=True)
